import json
import subprocess
import sys
import tempfile
import unittest
from pathlib import Path

repo_root = Path(__file__).resolve().parents[1]
sys.path.insert(0, str(repo_root))

from yolozu.adapter import RTDETRPoseAdapter  # noqa: E402
from yolozu.predictions import validate_predictions_entries, validate_predictions_payload  # noqa: E402


def _has_cuda() -> bool:
    try:
        import torch

        return bool(torch.cuda.is_available())
    except Exception:
        return False


def _write_tiny_rtdetr_pose_config(path: Path) -> None:
    cfg = {
        "dataset": {"root": ".", "split": "val2017", "format": "yolo"},
        "model": {
            "num_classes": 3,
            "hidden_dim": 64,
            "num_queries": 10,
            "use_uncertainty": False,
            "backbone_name": "tiny_cnn",
            "stem_channels": 8,
            "backbone_channels": [16, 32, 64],
            "stage_blocks": [1, 1, 1],
            "num_encoder_layers": 0,
            "num_decoder_layers": 1,
            "nhead": 8,
        },
        "loss": {"name": "default", "task_aligner": "none", "weights": {}},
        "train": {"batch_size": 1, "lr": 0.0001, "epochs": 1},
    }
    path.write_text(json.dumps(cfg))


@unittest.skipUnless(_has_cuda(), "CUDA not available")
class TestCudaSmokeRTDETRPose(unittest.TestCase):
    def test_adapter_predict_on_cuda(self):
        try:
            import torch
            from PIL import Image
        except ImportError as exc:  # pragma: no cover
            self.skipTest(str(exc))

        with tempfile.TemporaryDirectory() as td:
            root = Path(td)
            cfg_path = root / "tiny_rtdetr_pose.json"
            _write_tiny_rtdetr_pose_config(cfg_path)

            img_path = root / "img.jpg"
            Image.new("RGB", (8, 8), color=(0, 0, 0)).save(img_path)

            adapter = RTDETRPoseAdapter(
                config_path=str(cfg_path),
                checkpoint_path=None,
                device="cuda",
                image_size=(32, 32),
                score_threshold=1.1,  # force empty detections for stability
                max_detections=5,
            )
            model = adapter.get_model()
            param = next(model.parameters())
            self.assertTrue(param.is_cuda)

            outputs = adapter.predict([{"image": str(img_path)}])
            self.assertEqual(len(outputs), 1)
            self.assertEqual(outputs[0]["image"], str(img_path))
            self.assertIsInstance(outputs[0].get("detections"), list)

            res = validate_predictions_entries(outputs, strict=False)
            self.assertIsInstance(res.warnings, list)

            torch.cuda.synchronize()

    def test_export_predictions_cli_rtdetr_pose_cuda_smoke(self):
        try:
            from PIL import Image
        except ImportError as exc:  # pragma: no cover
            self.skipTest(str(exc))

        with tempfile.TemporaryDirectory(dir=str(repo_root)) as td:
            root = Path(td)
            dataset_root = root / "dataset"
            images = dataset_root / "images" / "val2017"
            labels = dataset_root / "labels" / "val2017"
            images.mkdir(parents=True, exist_ok=True)
            labels.mkdir(parents=True, exist_ok=True)

            img_path = images / "000001.jpg"
            Image.new("RGB", (8, 8), color=(0, 0, 0)).save(img_path)
            (labels / "000001.txt").write_text("0 0.5 0.5 0.2 0.2\n")

            cfg_path = root / "tiny_rtdetr_pose.json"
            _write_tiny_rtdetr_pose_config(cfg_path)

            out_path = root / "preds.json"
            script = repo_root / "tools" / "export_predictions.py"

            proc = subprocess.run(
                [
                    sys.executable,
                    str(script),
                    "--adapter",
                    "rtdetr_pose",
                    "--dataset",
                    str(dataset_root),
                    "--config",
                    str(cfg_path),
                    "--device",
                    "cuda",
                    "--image-size",
                    "32",
                    "--score-threshold",
                    "1.1",  # force empty detections for stability
                    "--max-detections",
                    "5",
                    "--max-images",
                    "1",
                    "--wrap",
                    "--output",
                    str(out_path),
                ],
                cwd=str(repo_root),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
                check=False,
            )
            if proc.returncode != 0:
                self.fail(f"export_predictions.py failed:\n{proc.stdout}\n{proc.stderr}")

            payload = json.loads(out_path.read_text())
            res = validate_predictions_payload(payload, strict=False)
            self.assertIsInstance(res.warnings, list)

    def test_export_predictions_cli_rtdetr_pose_cuda_smoke_with_lora(self):
        try:
            from PIL import Image
        except ImportError as exc:  # pragma: no cover
            self.skipTest(str(exc))

        with tempfile.TemporaryDirectory(dir=str(repo_root)) as td:
            root = Path(td)
            dataset_root = root / "dataset"
            images = dataset_root / "images" / "val2017"
            labels = dataset_root / "labels" / "val2017"
            images.mkdir(parents=True, exist_ok=True)
            labels.mkdir(parents=True, exist_ok=True)

            img_path = images / "000001.jpg"
            Image.new("RGB", (8, 8), color=(0, 0, 0)).save(img_path)
            (labels / "000001.txt").write_text("0 0.5 0.5 0.2 0.2\n")

            cfg_path = root / "tiny_rtdetr_pose.json"
            _write_tiny_rtdetr_pose_config(cfg_path)

            out_path = root / "preds.json"
            script = repo_root / "tools" / "export_predictions.py"

            proc = subprocess.run(
                [
                    sys.executable,
                    str(script),
                    "--adapter",
                    "rtdetr_pose",
                    "--dataset",
                    str(dataset_root),
                    "--config",
                    str(cfg_path),
                    "--device",
                    "cuda",
                    "--image-size",
                    "32",
                    "--score-threshold",
                    "1.1",  # force empty detections for stability
                    "--max-detections",
                    "5",
                    "--max-images",
                    "1",
                    "--wrap",
                    "--lora-r",
                    "2",
                    "--output",
                    str(out_path),
                ],
                cwd=str(repo_root),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
                check=False,
            )
            if proc.returncode != 0:
                self.fail(f"export_predictions.py failed:\n{proc.stdout}\n{proc.stderr}")

            payload = json.loads(out_path.read_text())
            meta = payload.get("meta") or {}
            self.assertTrue(meta.get("lora", {}).get("enabled"))
            report = meta.get("lora", {}).get("report")
            self.assertIsInstance(report, dict)
            self.assertTrue(report.get("enabled"))
            self.assertGreater(int(report.get("replaced", 0)), 0)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import csv
import json
import time
from pathlib import Path
from typing import Any


def now_utc_iso() -> str:
    return time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime())


def flatten_metrics(data: dict[str, Any], *, prefix: str = "", sep: str = ".") -> dict[str, Any]:
    out: dict[str, Any] = {}
    for key, value in (data or {}).items():
        k = f"{prefix}{sep}{key}" if prefix else str(key)
        if isinstance(value, dict):
            out.update(flatten_metrics(value, prefix=k, sep=sep))
        else:
            out[k] = value
    return out


def build_report(*, losses: dict[str, Any] | None = None, metrics: dict[str, Any] | None = None, meta: dict[str, Any] | None = None) -> dict[str, Any]:
    return {
        "schema_version": 1,
        "timestamp": now_utc_iso(),
        "losses": losses or {},
        "metrics": metrics or {},
        "meta": meta or {},
    }


def append_jsonl(path: str | Path, obj: dict[str, Any]) -> None:
    path = Path(path)
    path.parent.mkdir(parents=True, exist_ok=True)
    with path.open("a", encoding="utf-8") as f:
        f.write(json.dumps(obj, sort_keys=True) + "\n")


def write_json(path: str | Path, obj: dict[str, Any]) -> None:
    path = Path(path)
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(json.dumps(obj, indent=2, sort_keys=True))


def write_csv_row(path: str | Path, obj: dict[str, Any]) -> None:
    """Write a single-row CSV with stable flattened columns."""

    path = Path(path)
    path.parent.mkdir(parents=True, exist_ok=True)

    flat = flatten_metrics(obj)
    keys = sorted(flat.keys())
    with path.open("w", newline="", encoding="utf-8") as f:
        w = csv.DictWriter(f, fieldnames=keys)
        w.writeheader()
        w.writerow({k: flat.get(k) for k in keys})




"""Hessian-based solver for refining regression head predictions.

This module provides per-image iterative refinement of regression outputs
(depth, rotation, offsets) using second-order optimization (Gauss-Newton).
Unlike the L-BFGS calibration which optimizes dataset-wide scale and intrinsics,
this solver refines individual predictions using available geometric constraints.
"""

from __future__ import annotations

import math
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:  # pragma: no cover
    import torch


@dataclass(frozen=True)
class HessianSolverConfig:
    """Configuration for Hessian-based regression refinement.
    
    Attributes:
        max_iterations: Maximum number of Gauss-Newton iterations.
        convergence_threshold: Stop when parameter update norm is below this.
        damping: Levenberg-Marquardt damping factor (0 = pure Gauss-Newton).
        refine_depth: Whether to refine log_z predictions.
        refine_rotation: Whether to refine rot6d predictions.
        refine_offsets: Whether to refine center offsets.
        use_geometric_constraints: Use plane/upright constraints if available.
        w_depth: Weight for depth residuals.
        w_rotation: Weight for rotation residuals.
        w_offsets: Weight for offset residuals.
        w_plane: Weight for plane constraint (if enabled).
        w_upright: Weight for upright constraint (if enabled).
    """
    
    max_iterations: int = 5
    convergence_threshold: float = 1e-4
    damping: float = 1e-3
    refine_depth: bool = True
    refine_rotation: bool = True
    refine_offsets: bool = True
    use_geometric_constraints: bool = False
    w_depth: float = 1.0
    w_rotation: float = 1.0
    w_offsets: float = 1.0
    w_plane: float = 0.1
    w_upright: float = 0.1


@dataclass(frozen=True)
class RefinementResult:
    """Result of Hessian refinement for a single detection.
    
    Attributes:
        log_z: Refined log depth (or original if not refined).
        rot6d: Refined rotation 6D representation (or original if not refined).
        offsets: Refined center offsets (or original if not refined).
        iterations: Number of iterations performed.
        converged: Whether the solver converged.
        final_residual: Final residual norm.
    """
    
    log_z: float | None
    rot6d: list[float] | None
    offsets: list[float] | None
    iterations: int
    converged: bool
    final_residual: float


def _rot6d_to_matrix_torch(rot6d: torch.Tensor) -> torch.Tensor:
    """Convert 6D rotation representation to 3x3 rotation matrix.
    
    Args:
        rot6d: [..., 6] tensor of 6D rotation representation.
        
    Returns:
        [..., 3, 3] rotation matrix.
    """
    import torch
    from torch.nn import functional as F
    
    a1 = rot6d[..., 0:3]
    a2 = rot6d[..., 3:6]
    b1 = F.normalize(a1, dim=-1)
    b2 = F.normalize(a2 - (b1 * a2).sum(-1, keepdim=True) * b1, dim=-1)
    b3 = torch.cross(b1, b2, dim=-1)
    return torch.stack((b1, b2, b3), dim=-2)


def _geodesic_distance_torch(r1: torch.Tensor, r2: torch.Tensor) -> torch.Tensor:
    """Compute geodesic distance between rotation matrices.
    
    Args:
        r1: [..., 3, 3] rotation matrix.
        r2: [..., 3, 3] rotation matrix.
        
    Returns:
        [...] geodesic distance in radians.
    """
    import torch
    
    rel = torch.matmul(r1.transpose(-1, -2), r2)
    trace = rel[..., 0, 0] + rel[..., 1, 1] + rel[..., 2, 2]
    cos_theta = torch.clamp((trace - 1.0) * 0.5, -1.0, 1.0)
    return torch.acos(cos_theta)


def _basename_key(path_like: Any) -> str:
    text = str(path_like)
    if not text:
        return ""
    return text.replace("\\", "/").rsplit("/", 1)[-1]


def refine_detection_hessian(
    detection: dict[str, Any],
    *,
    config: HessianSolverConfig,
    gt_depth: float | None = None,
    gt_rotation: list[list[float]] | None = None,
    plane_constraint: dict[str, Any] | None = None,
    upright_bounds: tuple[tuple[float, float], tuple[float, float]] | None = None,
) -> dict[str, Any]:
    """Refine a single detection using Gauss-Newton with Hessian approximation.
    
    This function performs iterative refinement of regression outputs using
    available supervision signals and geometric constraints. It uses the
    Gauss-Newton method with optional Levenberg-Marquardt damping.
    
    Args:
        detection: Detection dictionary with 'log_z', 'rot6d', 'offsets' fields.
        config: Solver configuration.
        gt_depth: Optional ground truth depth for supervision.
        gt_rotation: Optional ground truth rotation matrix (3x3) for supervision.
        plane_constraint: Optional plane constraint {'n': [nx, ny, nz], 'd': d}.
        upright_bounds: Optional upright bounds ((roll_min, roll_max), (pitch_min, pitch_max)).
        
    Returns:
        Updated detection dictionary with refined predictions and metadata.
    """
    try:
        import torch
    except ImportError:  # pragma: no cover
        raise RuntimeError("torch is required for Hessian refinement")
    
    device = torch.device("cpu")
    dtype = torch.float32
    
    # Extract current predictions.
    log_z_val = detection.get("log_z")
    rot6d_val = detection.get("rot6d")
    offsets_val = detection.get("offsets")
    
    # Initialize parameters to refine.
    params = []
    param_types = []
    
    if config.refine_depth and log_z_val is not None:
        log_z_param = torch.tensor([float(log_z_val)], dtype=dtype, device=device, requires_grad=True)
        params.append(log_z_param)
        param_types.append("log_z")
    else:
        log_z_param = None
        
    if config.refine_rotation and rot6d_val is not None:
        try:
            rot6d_list = [float(x) for x in rot6d_val]
            if len(rot6d_list) == 6:
                rot6d_param = torch.tensor(rot6d_list, dtype=dtype, device=device, requires_grad=True)
                params.append(rot6d_param)
                param_types.append("rot6d")
            else:
                rot6d_param = None
        except (TypeError, ValueError):
            rot6d_param = None
    else:
        rot6d_param = None
        
    if config.refine_offsets and offsets_val is not None:
        try:
            offsets_list = [float(x) for x in offsets_val]
            if len(offsets_list) == 2:
                offsets_param = torch.tensor(offsets_list, dtype=dtype, device=device, requires_grad=True)
                params.append(offsets_param)
                param_types.append("offsets")
            else:
                offsets_param = None
        except (TypeError, ValueError):
            offsets_param = None
    else:
        offsets_param = None
    
    # If nothing to refine, return original.
    if not params:
        return detection
    
    # Build supervision tensors.
    z_gt_tensor = None
    if gt_depth is not None:
        z_gt_tensor = torch.tensor(float(gt_depth), dtype=dtype, device=device)
        
    r_gt_tensor = None
    if gt_rotation is not None:
        try:
            r_flat = [float(x) for row in gt_rotation for x in row]
            if len(r_flat) == 9:
                r_gt_tensor = torch.tensor(r_flat, dtype=dtype, device=device).reshape(3, 3)
        except (TypeError, ValueError):
            pass
    
    # Gauss-Newton iterations.
    converged = False
    final_residual = float("inf")
    iterations_run = 0
    
    for iteration in range(config.max_iterations):
        iterations_run = int(iteration + 1)
        # Compute residuals.
        residuals = []
        weights = []
        
        # Depth residual.
        if log_z_param is not None and z_gt_tensor is not None:
            z_pred = torch.exp(log_z_param[0]).clamp(min=1e-6)
            r_depth = z_pred - z_gt_tensor
            residuals.append(r_depth)
            weights.append(config.w_depth)
        
        # Rotation residual.
        if rot6d_param is not None and r_gt_tensor is not None:
            r_pred = _rot6d_to_matrix_torch(rot6d_param.unsqueeze(0)).squeeze(0)
            geo_dist = _geodesic_distance_torch(r_pred.unsqueeze(0), r_gt_tensor.unsqueeze(0)).squeeze(0)
            residuals.append(geo_dist)
            weights.append(config.w_rotation)
        
        # Offset residual (regularization to keep offsets small).
        if offsets_param is not None:
            r_offsets = (offsets_param * offsets_param).sum().sqrt()
            residuals.append(r_offsets)
            weights.append(config.w_offsets * 0.01)  # Weak regularization.
        
        # Plane constraint (if using geometric constraints).
        if config.use_geometric_constraints and plane_constraint is not None:
            # Plane constraint would need translation which requires intrinsics.
            # Skip for now as this is a simplified implementation.
            pass
        
        # Upright constraint.
        if config.use_geometric_constraints and upright_bounds is not None and rot6d_param is not None:
            r_pred = _rot6d_to_matrix_torch(rot6d_param.unsqueeze(0)).squeeze(0)
            roll = torch.atan2(r_pred[2, 1], r_pred[2, 2])
            pitch = torch.asin(torch.clamp(-r_pred[2, 0], -1.0, 1.0))
            
            (roll_min, roll_max), (pitch_min, pitch_max) = upright_bounds
            roll_penalty = torch.clamp(roll - roll_max, min=0) + torch.clamp(roll_min - roll, min=0)
            pitch_penalty = torch.clamp(pitch - pitch_max, min=0) + torch.clamp(pitch_min - pitch, min=0)
            r_upright = roll_penalty + pitch_penalty
            residuals.append(r_upright)
            weights.append(config.w_upright)
        
        if not residuals:
            final_residual = 0.0
            break
        
        n_params = int(sum(int(p.numel()) for p in params))
        
        # Ensure all residuals are scalar and weighted.
        scalar_residuals = []
        for r, w in zip(residuals, weights):
            # Ensure residual is scalar.
            if r.numel() > 1:
                r_scalar = r.mean()  # Reduce to scalar if needed.
            else:
                r_scalar = r.squeeze()
            scalar_residuals.append(w * r_scalar)
        
        # Total loss for convergence check.
        total_residual = sum(r * r for r in scalar_residuals)
        final_residual = float(total_residual.sqrt().item())
        
        # Compute gradients for each residual.
        J_rows = []
        for r in scalar_residuals:
            # Compute gradient of this residual w.r.t. all parameters.
            grads = torch.autograd.grad(r, params, retain_graph=True, create_graph=False, allow_unused=True)
            grad_row = torch.cat([g.flatten() if g is not None else torch.zeros_like(p).flatten() 
                                 for g, p in zip(grads, params)])
            J_rows.append(grad_row)
        
        if not J_rows:
            break
        
        J = torch.stack(J_rows, dim=0)  # [n_residuals, n_params]
        r_vec = torch.stack([r for r in scalar_residuals], dim=0)  # [n_residuals]
        
        # Gauss-Newton with Levenberg-Marquardt damping:
        # (J^T J + λI) Δp = -J^T r
        JtJ = torch.matmul(J.t(), J)
        Jtr = torch.matmul(J.t(), r_vec)
        
        # Add damping.
        damping_matrix = config.damping * torch.eye(n_params, dtype=dtype, device=device)
        A = JtJ + damping_matrix
        
        # Solve for parameter update.
        try:
            delta_p = torch.linalg.solve(A, -Jtr)
        except RuntimeError:
            # Singular matrix, stop.
            break
        
        # Check convergence.
        update_norm = delta_p.norm().item()
        
        # Check for NaN.
        if math.isnan(update_norm) or math.isnan(final_residual):
            # Numerical instability, stop refinement.
            break
        
        if update_norm < config.convergence_threshold:
            converged = True
            break
        
        # Update parameters (detach and re-create to avoid autograd graph explosion).
        offset = 0
        new_params = []
        for i, param in enumerate(params):
            size = param.numel()
            delta = delta_p[offset:offset + size].reshape(param.shape)
            updated = param.data + delta
            new_param = updated.clone().detach().requires_grad_(True)
            new_params.append(new_param)
            offset += size
        
        params = new_params
        # Re-assign to individual param variables.
        param_idx = 0
        if "log_z" in param_types:
            log_z_param = params[param_idx]
            param_idx += 1
        if "rot6d" in param_types:
            rot6d_param = params[param_idx]
            param_idx += 1
        if "offsets" in param_types:
            offsets_param = params[param_idx]
            param_idx += 1
    
    # Extract refined values.
    refined_log_z = None
    if log_z_param is not None:
        refined_log_z = float(log_z_param[0].detach().cpu().item())
    
    refined_rot6d = None
    if rot6d_param is not None:
        refined_rot6d = [float(x) for x in rot6d_param.detach().cpu().tolist()]
    
    refined_offsets = None
    if offsets_param is not None:
        refined_offsets = [float(x) for x in offsets_param.detach().cpu().tolist()]
    
    # Build result.
    result = RefinementResult(
        log_z=refined_log_z,
        rot6d=refined_rot6d,
        offsets=refined_offsets,
        iterations=int(iterations_run),
        converged=converged,
        final_residual=final_residual,
    )
    
    # Update detection with refined values.
    refined_detection = dict(detection)
    if refined_log_z is not None:
        refined_detection["log_z"] = refined_log_z
    if refined_rot6d is not None:
        refined_detection["rot6d"] = refined_rot6d
    if refined_offsets is not None:
        refined_detection["offsets"] = refined_offsets
    
    # Add refinement metadata.
    refined_detection["hessian_refinement"] = {
        "iterations": result.iterations,
        "converged": result.converged,
        "final_residual": result.final_residual,
    }
    
    return refined_detection


def refine_predictions_hessian(
    predictions: list[dict[str, Any]],
    records: list[dict[str, Any]] | None = None,
    *,
    config: HessianSolverConfig,
) -> list[dict[str, Any]]:
    """Refine predictions using Hessian-based optimization.
    
    This function applies per-detection refinement to all predictions.
    If records are provided, uses available ground truth for supervision.
    
    Args:
        predictions: List of prediction dictionaries with 'image' and 'detections'.
        records: Optional list of dataset records with ground truth.
        config: Solver configuration.
        
    Returns:
        List of refined predictions.
    """
    # Build record index if provided.
    record_index: dict[str, dict[str, Any]] = {}
    if records is not None:
        for record in records:
            if not isinstance(record, dict):
                continue
            image = record.get("image")
            if image:
                record_index[str(image)] = record
                # Also index by basename.
                base = _basename_key(image)
                if base and base not in record_index:
                    record_index[base] = record
    
    refined_predictions = []
    
    for pred_entry in predictions:
        if not isinstance(pred_entry, dict):
            refined_predictions.append(pred_entry)
            continue
        
        image = pred_entry.get("image")
        dets = pred_entry.get("detections")
        if not isinstance(dets, list):
            refined_predictions.append(pred_entry)
            continue
        
        # Find corresponding record if available.
        record = None
        if image and record_index:
            record = record_index.get(str(image))
            if record is None:
                base = _basename_key(image)
                record = record_index.get(base)
        
        # Refine each detection.
        refined_dets = []
        for det in dets:
            if not isinstance(det, dict):
                refined_dets.append(det)
                continue
            
            # Extract supervision from record if available.
            # NOTE: This is a simplified matching strategy that uses the first GT instance.
            # For production use with multi-object scenes, implement proper detection-to-GT
            # matching (e.g., IoU-based Hungarian matching or class_id + spatial proximity).
            gt_depth = None
            gt_rotation = None
            
            if record is not None:
                labels = record.get("labels")
                if isinstance(labels, list) and labels:
                    # Try to match by class_id first, otherwise use first label.
                    matched_label = None
                    det_class = det.get("class_id")
                    
                    if det_class is not None:
                        # Find first label with matching class_id.
                        for label in labels:
                            if isinstance(label, dict) and label.get("class_id") == det_class:
                                matched_label = label
                                break
                    
                    # Fallback to first label if no class match.
                    if matched_label is None and labels:
                        matched_label = labels[0]
                    
                    if matched_label is not None and isinstance(matched_label, dict):
                        # Try to extract depth from t_gt.
                        t_gt = matched_label.get("t_gt")
                        if isinstance(t_gt, (list, tuple)) and len(t_gt) >= 3:
                            try:
                                gt_depth = float(t_gt[2])  # Z component.
                            except (TypeError, ValueError):
                                pass
                        
                        # Extract rotation.
                        r_gt = matched_label.get("R_gt")
                        if isinstance(r_gt, (list, tuple)) and len(r_gt) == 3:
                            try:
                                gt_rotation = [[float(x) for x in row] for row in r_gt]
                            except (TypeError, ValueError):
                                pass
            
            # Refine detection.
            refined_det = refine_detection_hessian(
                det,
                config=config,
                gt_depth=gt_depth,
                gt_rotation=gt_rotation,
            )
            refined_dets.append(refined_det)
        
        # Build refined entry.
        refined_entry = dict(pred_entry)
        refined_entry["detections"] = refined_dets
        refined_predictions.append(refined_entry)
    
    return refined_predictions



from __future__ import annotations

import json
import random
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Iterable

from .gates import final_score, passes_template_gate


@dataclass(frozen=True)
class TransformResult:
    entries: list[dict[str, Any]]
    warnings: list[str]

def load_classes_json(path: str | Path) -> dict[str, Any]:
    return json.loads(Path(path).read_text())


def _to_int_key_map(mapping: dict[Any, Any]) -> dict[int, int]:
    out: dict[int, int] = {}
    for k, v in mapping.items():
        try:
            out[int(k)] = int(v)
        except Exception:
            continue
    return out


def build_category_id_to_class_id_map(classes_json: dict[str, Any]) -> dict[int, int]:
    raw = classes_json.get("category_id_to_class_id")
    if not isinstance(raw, dict):
        raise ValueError("classes.json missing category_id_to_class_id")
    return _to_int_key_map(raw)


def normalize_class_ids(
    entries: Iterable[dict[str, Any]],
    *,
    classes_json: dict[str, Any] | None = None,
    assume_class_id_is_category_id: bool = False,
) -> TransformResult:
    """Normalize detections to use contiguous `class_id` (0..N-1).

    Supported normalization:
    - If a detection has `category_id` and is missing `class_id`, map it.
    - If assume_class_id_is_category_id=True, treat `class_id` as a COCO category id and map it.
    """

    warnings: list[str] = []
    cat_to_cls: dict[int, int] | None = None
    if classes_json is not None:
        cat_to_cls = build_category_id_to_class_id_map(classes_json)

    out_entries: list[dict[str, Any]] = []
    for idx, entry in enumerate(entries):
        if not isinstance(entry, dict):
            continue
        new_entry = dict(entry)
        dets = new_entry.get("detections") or []
        if not isinstance(dets, list):
            dets = []

        new_dets = []
        for j, det in enumerate(dets):
            if not isinstance(det, dict):
                continue
            new_det = dict(det)

            if assume_class_id_is_category_id and "class_id" in new_det and cat_to_cls is not None:
                try:
                    cat_id = int(new_det["class_id"])
                    if cat_id in cat_to_cls:
                        new_det["class_id"] = int(cat_to_cls[cat_id])
                    else:
                        warnings.append(f"predictions[{idx}].detections[{j}]: unknown category_id {cat_id}")
                except Exception:
                    warnings.append(f"predictions[{idx}].detections[{j}]: invalid class_id")

            if "class_id" not in new_det and "category_id" in new_det and cat_to_cls is not None:
                try:
                    cat_id = int(new_det["category_id"])
                    if cat_id in cat_to_cls:
                        new_det["class_id"] = int(cat_to_cls[cat_id])
                    else:
                        warnings.append(f"predictions[{idx}].detections[{j}]: unknown category_id {cat_id}")
                except Exception:
                    warnings.append(f"predictions[{idx}].detections[{j}]: invalid category_id")

            new_dets.append(new_det)

        new_entry["detections"] = new_dets
        out_entries.append(new_entry)

    return TransformResult(entries=out_entries, warnings=warnings)


def _entry_image_size(entry: dict[str, Any]) -> tuple[float | None, float | None]:
    value = entry.get("image_size")
    if isinstance(value, dict):
        width = value.get("width")
        height = value.get("height")
        try:
            return (float(width), float(height))
        except Exception:
            return (None, None)
    if isinstance(value, (list, tuple)) and len(value) == 2:
        try:
            return (float(value[0]), float(value[1]))
        except Exception:
            return (None, None)
    return (None, None)


def apply_tta(
    entries: Iterable[dict[str, Any]],
    *,
    enabled: bool = True,
    seed: int | None = None,
    flip_prob: float = 0.5,
    norm_only: bool = False,
) -> TransformResult:
    """Apply a simple test-time augmentation transform to predictions.

    When enabled, a per-detection mask is sampled (seeded) to decide whether
    to apply a horizontal flip in normalized space (cx -> 1 - cx). If
    norm_only is False and bbox_abs is present, a best-effort absolute flip is
    applied using entry.image_size.
    """

    warnings: list[str] = []
    out_entries: list[dict[str, Any]] = []
    rng = random.Random(seed)

    for idx, entry in enumerate(entries):
        if not isinstance(entry, dict):
            continue
        new_entry = dict(entry)
        dets = new_entry.get("detections") or []
        if not isinstance(dets, list):
            dets = []

        new_dets = []
        mask: list[bool] = []

        for j, det in enumerate(dets):
            if not isinstance(det, dict):
                continue
            new_det = dict(det)
            apply = bool(enabled) and (rng.random() < float(flip_prob))
            mask.append(apply)

            if apply:
                bbox = new_det.get("bbox")
                if isinstance(bbox, dict) and all(k in bbox for k in ("cx", "cy", "w", "h")):
                    new_bbox = dict(bbox)
                    try:
                        new_bbox["cx"] = 1.0 - float(new_bbox["cx"])
                    except Exception:
                        warnings.append(f"predictions[{idx}].detections[{j}]: invalid bbox.cx")
                    new_det["bbox"] = new_bbox
                else:
                    warnings.append(f"predictions[{idx}].detections[{j}]: missing bbox for tta")

                if not norm_only:
                    bbox_abs = new_det.get("bbox_abs")
                    if isinstance(bbox_abs, dict) and "cx" in bbox_abs:
                        width, _ = _entry_image_size(entry)
                        if width is None:
                            warnings.append(f"predictions[{idx}].detections[{j}]: missing image_size for bbox_abs")
                        else:
                            new_bbox_abs = dict(bbox_abs)
                            try:
                                new_bbox_abs["cx"] = float(width) - float(new_bbox_abs["cx"])
                            except Exception:
                                warnings.append(f"predictions[{idx}].detections[{j}]: invalid bbox_abs.cx")
                            new_det["bbox_abs"] = new_bbox_abs

            new_dets.append(new_det)

        new_entry["detections"] = new_dets
        if enabled:
            new_entry["tta_mask"] = mask
        out_entries.append(new_entry)

    return TransformResult(entries=out_entries, warnings=warnings)

def fuse_detection_scores(
    entries: Iterable[dict[str, Any]],
    *,
    weights: dict[str, float] | None = None,
    det_score_key: str = "score",
    template_score_key: str = "score_tmp_sym",
    sigma_z_key: str = "sigma_z",
    sigma_rot_key: str = "sigma_rot",
    out_score_key: str = "score",
    preserve_det_score_key: str | None = "score_det",
    template_gate_enabled: bool = False,
    template_gate_tau: float = 0.0,
    min_score: float | None = None,
    topk_per_image: int | None = None,
) -> TransformResult:
    """Fuse detection scores for inference-time gating/tuning.

    The fused score is:
      w_det * score_det + w_tmp * score_tmp_sym - w_unc * (sigma_z + sigma_rot)

    This is intended for *postprocess-time* score shaping (ordering/thresholding),
    and can be tuned offline on a fixed eval subset.

    Notes:
    - Missing template/uncertainty fields default to 0.0.
    - If template_gate_enabled=True, detections with score_tmp_sym < template_gate_tau are dropped.
    - If preserve_det_score_key is set and missing, the original det score is stored there.
    """

    warnings: list[str] = []
    w = {"det": 1.0, "tmp": 1.0, "unc": 1.0}
    if weights:
        for k, v in weights.items():
            try:
                w[str(k)] = float(v)
            except Exception:
                continue

    out_entries: list[dict[str, Any]] = []
    for idx, entry in enumerate(entries):
        if not isinstance(entry, dict):
            continue
        image = entry.get("image")
        if not image:
            warnings.append(f"predictions[{idx}]: missing image")
            continue

        dets = entry.get("detections") or []
        if not isinstance(dets, list):
            warnings.append(f"predictions[{idx}]: detections must be a list")
            dets = []

        new_dets: list[dict[str, Any]] = []
        for j, det in enumerate(dets):
            if not isinstance(det, dict):
                continue

            try:
                score_det = float(det.get(det_score_key, 0.0))
            except Exception:
                warnings.append(f"predictions[{idx}].detections[{j}]: invalid {det_score_key}")
                score_det = 0.0

            try:
                score_tmp = float(det.get(template_score_key, 0.0))
            except Exception:
                warnings.append(f"predictions[{idx}].detections[{j}]: invalid {template_score_key}")
                score_tmp = 0.0

            try:
                sigma_z = float(det.get(sigma_z_key, 0.0))
            except Exception:
                warnings.append(f"predictions[{idx}].detections[{j}]: invalid {sigma_z_key}")
                sigma_z = 0.0

            try:
                sigma_rot = float(det.get(sigma_rot_key, 0.0))
            except Exception:
                warnings.append(f"predictions[{idx}].detections[{j}]: invalid {sigma_rot_key}")
                sigma_rot = 0.0

            if template_gate_enabled and not passes_template_gate(score_tmp, enabled=True, tau=float(template_gate_tau)):
                continue

            fused = final_score(score_det, score_tmp, sigma_z, sigma_rot, w)
            if min_score is not None and fused < float(min_score):
                continue

            new_det = dict(det)
            if preserve_det_score_key and preserve_det_score_key not in new_det:
                new_det[preserve_det_score_key] = float(score_det)
            new_det[out_score_key] = float(fused)
            new_dets.append(new_det)

        if topk_per_image is not None and topk_per_image > 0 and len(new_dets) > topk_per_image:
            new_dets.sort(key=lambda d: float(d.get(out_score_key, 0.0)), reverse=True)
            new_dets = new_dets[: int(topk_per_image)]

        out_entries.append({"image": image, "detections": new_dets})

    return TransformResult(entries=out_entries, warnings=warnings)



import json
import sys
import tempfile
import unittest
from pathlib import Path

sys.path.insert(0, str(Path(__file__).resolve().parents[1]))

from yolozu.adapter import PrecomputedAdapter


class TestPrecomputedAdapter(unittest.TestCase):
    def test_matches_by_full_path(self):
        with tempfile.TemporaryDirectory() as td:
            path = Path(td) / "preds.json"
            payload = [
                {"image": "/tmp/a.jpg", "detections": [{"score": 0.9}]},
            ]
            path.write_text(json.dumps(payload))
            adapter = PrecomputedAdapter(path)
            out = adapter.predict([{"image": "/tmp/a.jpg", "labels": []}])
            self.assertEqual(out[0]["detections"], [{"score": 0.9}])

    def test_matches_by_basename(self):
        with tempfile.TemporaryDirectory() as td:
            path = Path(td) / "preds.json"
            payload = {"0001.jpg": [{"score": 0.5}]}
            path.write_text(json.dumps(payload))
            adapter = PrecomputedAdapter(path)
            out = adapter.predict([{"image": "/any/dir/0001.jpg", "labels": []}])
            self.assertEqual(out[0]["detections"], [{"score": 0.5}])

    def test_top_level_predictions_key(self):
        with tempfile.TemporaryDirectory() as td:
            path = Path(td) / "preds.json"
            payload = {"predictions": [{"image": "x.jpg", "detections": []}]}
            path.write_text(json.dumps(payload))
            adapter = PrecomputedAdapter(path)
            out = adapter.predict([{"image": "x.jpg", "labels": []}])
            self.assertEqual(out[0]["detections"], [])

    def test_matches_windows_style_path_alias(self):
        with tempfile.TemporaryDirectory() as td:
            path = Path(td) / "preds.json"
            payload = [{"image": r"C:\tmp\0001.jpg", "detections": [{"score": 0.8}]}]
            path.write_text(json.dumps(payload))
            adapter = PrecomputedAdapter(path)
            out = adapter.predict([{"image": "C:/tmp/0001.jpg", "labels": []}])
            self.assertEqual(out[0]["detections"], [{"score": 0.8}])


if __name__ == "__main__":
    unittest.main()



import sys
from pathlib import Path
import unittest

sys.path.insert(0, str(Path(__file__).resolve().parents[1]))

from yolozu.boxes import xyxy_abs_to_cxcywh_norm


class TestBoxes(unittest.TestCase):
    def test_xyxy_to_cxcywh_norm(self):
        # Image 100x200, bbox x:[10,30], y:[20,60]
        cx, cy, w, h = xyxy_abs_to_cxcywh_norm((10.0, 20.0, 30.0, 60.0), width=100, height=200)
        self.assertAlmostEqual(cx, 0.2, places=6)  # (10+30)/2 / 100
        self.assertAlmostEqual(cy, 0.2, places=6)  # (20+60)/2 / 200
        self.assertAlmostEqual(w, 0.2, places=6)   # 20/100
        self.assertAlmostEqual(h, 0.2, places=6)   # 40/200


if __name__ == "__main__":
    unittest.main()




import sys
import unittest
from pathlib import Path

sys.path.insert(0, str(Path(__file__).resolve().parents[1]))

from yolozu.tta import TTARunner, TTAConfig, apply_tta_transform


class TestTTABase(unittest.TestCase):
    def test_apply_tta_transform_disabled(self):
        entries = [{"image": "x.jpg", "detections": [{"bbox": {"cx": 0.2, "cy": 0.3, "w": 0.1, "h": 0.1}}]}]
        cfg = TTAConfig(enabled=False, seed=1)
        out = apply_tta_transform(entries, config=cfg)
        self.assertEqual(out.entries[0]["detections"][0]["bbox"]["cx"], 0.2)
        self.assertFalse(out.warnings)

    def test_apply_tta_transform_enabled(self):
        entries = [{"image": "x.jpg", "detections": [{"bbox": {"cx": 0.2, "cy": 0.3, "w": 0.1, "h": 0.1}}]}]
        cfg = TTAConfig(enabled=True, seed=0, flip_prob=1.0)
        out = apply_tta_transform(entries, config=cfg)
        self.assertAlmostEqual(out.entries[0]["detections"][0]["bbox"]["cx"], 0.8)

    def test_tta_runner_interface(self):
        class Dummy(TTARunner):
            def __init__(self):
                self.steps = 0

            def adapt_step(self, batch):
                self.steps += 1
                return {"loss": 0.0}

        runner = Dummy()
        runner.reset()
        out = runner.adapt_step({"x": 1})
        self.assertIn("loss", out)
        self.assertIsNone(runner.maybe_log())


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import math
from typing import Any, Iterable

from .constraints import apply_constraints
from .geometry import corrected_intrinsics, recover_translation


def _image_size(entry: dict[str, Any]) -> tuple[float | None, float | None]:
    if "image_hw" in entry:
        value = entry.get("image_hw")
        if isinstance(value, (list, tuple)) and len(value) == 2:
            return float(value[1]), float(value[0])
    value = entry.get("image_size")
    if isinstance(value, dict):
        try:
            return float(value.get("width")), float(value.get("height"))
        except Exception:
            return (None, None)
    if isinstance(value, (list, tuple)) and len(value) == 2:
        try:
            return float(value[0]), float(value[1])
        except Exception:
            return (None, None)
    return (None, None)


def _intrinsics_from_value(value: Any) -> tuple[float, float, float, float] | None:
    if value is None:
        return None
    if isinstance(value, dict):
        if all(k in value for k in ("fx", "fy", "cx", "cy")):
            return (float(value["fx"]), float(value["fy"]), float(value["cx"]), float(value["cy"]))
    if isinstance(value, (list, tuple)) and len(value) == 3:
        try:
            fx = float(value[0][0])
            fy = float(value[1][1])
            cx = float(value[0][2])
            cy = float(value[1][2])
            return (fx, fy, cx, cy)
        except Exception:
            return None
    return None


def _intrinsics(entry: dict[str, Any], det: dict[str, Any]) -> tuple[float, float, float, float] | None:
    for key in ("intrinsics", "K_gt", "K"):
        value = det.get(key)
        got = _intrinsics_from_value(value)
        if got is not None:
            return got
    for key in ("intrinsics", "K_gt", "K"):
        value = entry.get(key)
        got = _intrinsics_from_value(value)
        if got is not None:
            return got
    return None


def _rot6d_to_matrix(rot6d: Iterable[float]) -> list[list[float]] | None:
    vals = list(rot6d)
    if len(vals) != 6:
        return None
    a1 = vals[0:3]
    a2 = vals[3:6]

    def _norm(v):
        n = (v[0] ** 2 + v[1] ** 2 + v[2] ** 2) ** 0.5
        if n <= 0:
            return [0.0, 0.0, 0.0]
        return [v[0] / n, v[1] / n, v[2] / n]

    b1 = _norm(a1)
    dot = b1[0] * a2[0] + b1[1] * a2[1] + b1[2] * a2[2]
    a2o = [a2[0] - dot * b1[0], a2[1] - dot * b1[1], a2[2] - dot * b1[2]]
    b2 = _norm(a2o)
    b3 = [
        b1[1] * b2[2] - b1[2] * b2[1],
        b1[2] * b2[0] - b1[0] * b2[2],
        b1[0] * b2[1] - b1[1] * b2[0],
    ]
    return [b1, b2, b3]


def _bbox_center_wh(
    bbox: dict[str, Any],
    *,
    image_wh: tuple[float | None, float | None],
    bbox_format: str,
) -> tuple[tuple[float, float], tuple[float, float]] | None:
    if not all(k in bbox for k in ("cx", "cy", "w", "h")):
        return None
    cx = float(bbox["cx"])
    cy = float(bbox["cy"])
    bw = float(bbox["w"])
    bh = float(bbox["h"])
    if bbox_format == "cxcywh_norm":
        width, height = image_wh
        if width is None or height is None:
            return None
        return (cx * width, cy * height), (bw * width, bh * height)
    if bbox_format == "cxcywh_abs":
        return (cx, cy), (bw, bh)
    raise ValueError(f"unsupported bbox_format: {bbox_format}")


def _gate_ok(cfg: dict[str, Any], constraints: dict[str, Any]) -> bool:
    enabled = cfg.get("enabled", {}) if isinstance(cfg, dict) else {}
    if enabled.get("depth_prior", False) and constraints.get("depth_range_violation", 0.0) > 0:
        return False
    if enabled.get("table_plane", False) and not bool(constraints.get("plane_ok", True)):
        return False
    if enabled.get("upright", False) and constraints.get("upright_violation", 0.0) > 0:
        return False
    return True


def infer_constraints(
    entries: Iterable[dict[str, Any]],
    *,
    constraints_cfg: dict[str, Any],
    bbox_format: str = "cxcywh_norm",
    default_size_wh: tuple[float, float] = (1.0, 1.0),
) -> list[dict[str, Any]]:
    out_entries: list[dict[str, Any]] = []
    for entry in entries:
        if not isinstance(entry, dict):
            continue
        image_wh = _image_size(entry)
        new_entry = dict(entry)
        dets = new_entry.get("detections") or []
        if not isinstance(dets, list):
            dets = []
        new_dets = []
        for det in dets:
            if not isinstance(det, dict):
                continue
            new_det = dict(det)
            bbox = new_det.get("bbox") if isinstance(new_det.get("bbox"), dict) else None
            z_pred = None
            if "log_z" in new_det:
                try:
                    z_pred = float(math.exp(float(new_det["log_z"])))
                except Exception:
                    z_pred = None
            if z_pred is None and "z" in new_det:
                try:
                    z_pred = float(new_det["z"])
                except Exception:
                    z_pred = None

            offsets = new_det.get("offsets") or [0.0, 0.0]
            if not isinstance(offsets, (list, tuple)) or len(offsets) != 2:
                offsets = [0.0, 0.0]
            offsets = (float(offsets[0]), float(offsets[1]))

            k_delta = new_det.get("k_delta")
            if not isinstance(k_delta, (list, tuple)) or len(k_delta) != 4:
                k_delta = None

            intrinsics = _intrinsics(entry, new_det)
            bbox_info = _bbox_center_wh(bbox, image_wh=image_wh, bbox_format=bbox_format) if bbox else None
            size_wh = new_det.get("size_wh") or entry.get("size_wh") or default_size_wh
            if not isinstance(size_wh, (list, tuple)) or len(size_wh) != 2:
                size_wh = default_size_wh
            size_wh = (float(size_wh[0]), float(size_wh[1]))

            r_mat = new_det.get("R") or new_det.get("r_mat") or new_det.get("rot_mat")
            if r_mat is None and "rot6d" in new_det:
                r_mat = _rot6d_to_matrix(new_det.get("rot6d"))

            if intrinsics and bbox_info and z_pred is not None:
                k_prime = corrected_intrinsics(intrinsics, k_delta) if k_delta is not None else intrinsics
                t_xyz = recover_translation(bbox_info[0], offsets, z_pred, k_prime)
                new_det["k_prime"] = list(k_prime)
                new_det["t_xyz"] = list(t_xyz)
                class_key = new_det.get("class_id", new_det.get("class_name"))
                constraints = apply_constraints(
                    constraints_cfg,
                    class_key=class_key,
                    bbox_wh=bbox_info[1],
                    size_wh=size_wh,
                    intrinsics_fx_fy=(k_prime[0], k_prime[1]),
                    t_xyz=t_xyz,
                    r_mat=r_mat or [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]],
                    z_pred=z_pred,
                )
                new_det["constraints"] = constraints
                new_det["gate_ok"] = _gate_ok(constraints_cfg, constraints)
            new_dets.append(new_det)

        new_entry["detections"] = new_dets
        out_entries.append(new_entry)
    return out_entries



import sys
import unittest
from pathlib import Path

sys.path.insert(0, str(Path(__file__).resolve().parents[1]))

from tools import build_trt_engine


class _DummyInput:
    def __init__(self, name):
        self.name = name


class _DummyNetwork:
    def __init__(self, names):
        self._inputs = [_DummyInput(n) for n in names]

    @property
    def num_inputs(self):
        return len(self._inputs)

    def get_input(self, idx):
        return self._inputs[idx]


class TestBuildTrtEngine(unittest.TestCase):
    def test_resolve_input_name_prefers_requested(self):
        net = _DummyNetwork(["images", "other"])
        name = build_trt_engine._resolve_input_name(net, "images")
        self.assertEqual(name, "images")

    def test_resolve_input_name_falls_back(self):
        net = _DummyNetwork(["data"])
        name = build_trt_engine._resolve_input_name(net, "images")
        self.assertEqual(name, "data")

    def test_resolve_input_name_no_inputs(self):
        net = _DummyNetwork([])
        name = build_trt_engine._resolve_input_name(net, "images")
        self.assertEqual(name, "images")


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import platform
import subprocess
import sys
from pathlib import Path
from typing import Any


def _safe_version(module_name: str) -> str | None:
    try:
        mod = __import__(module_name)
    except Exception:
        return None
    version = getattr(mod, "__version__", None)
    if version is None:
        return None
    # Some libraries (notably PyTorch) use a custom string subclass for __version__.
    # Cast to a plain str so it remains safe/portable when saved in torch checkpoints.
    return str(version)


def git_info(repo_root: str | Path) -> dict[str, Any]:
    """Return best-effort git metadata for repo_root.

    Never raises; returns empty dict if git is unavailable or repo_root is not a git repo.
    """

    root = Path(repo_root)
    try:
        sha = subprocess.check_output(
            ["git", "-C", str(root), "rev-parse", "HEAD"],
            stderr=subprocess.DEVNULL,
            text=True,
        ).strip()
        dirty = subprocess.call(
            ["git", "-C", str(root), "diff", "--quiet"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        # diff --quiet returns 1 when there are changes
        is_dirty = bool(dirty != 0)
        return {"sha": sha, "dirty": is_dirty}
    except Exception:
        return {}


def versions() -> dict[str, Any]:
    return {
        "python": platform.python_version(),
        "platform": platform.platform(),
        "numpy": _safe_version("numpy"),
        "torch": _safe_version("torch"),
        "Pillow": _safe_version("PIL"),
        "PyYAML": _safe_version("yaml"),
    }


def build_run_record(
    *,
    repo_root: str | Path,
    argv: list[str] | None = None,
    args: dict[str, Any] | None = None,
    dataset_root: str | None = None,
    extra: dict[str, Any] | None = None,
) -> dict[str, Any]:
    """Build a reproducibility record for logs/checkpoints.

    This is intentionally best-effort (never raises) so it can run in CI and non-git envs.
    """

    record: dict[str, Any] = {
        "versions": versions(),
        "git": git_info(repo_root),
    }

    if argv is None:
        argv = sys.argv[1:]
    record["argv"] = list(argv)

    if args is not None:
        record["args"] = dict(args)

    if dataset_root is not None:
        record["dataset_root"] = str(dataset_root)

    if extra:
        record["extra"] = dict(extra)

    return record



import tempfile
import unittest
from pathlib import Path


def _write_stub_png(path: Path, *, width: int, height: int) -> None:
    # Minimal PNG header sufficient for yolozu.image_size.get_image_size (no CRC validation).
    sig = b"\x89PNG\r\n\x1a\n"
    ihdr_len = (13).to_bytes(4, "big", signed=False)
    ihdr = b"IHDR"
    w = int(width).to_bytes(4, "big", signed=False)
    h = int(height).to_bytes(4, "big", signed=False)
    # bitdepth=8, colortype=2, compression=0, filter=0, interlace=0
    rest = bytes([8, 2, 0, 0, 0])
    path.write_bytes(sig + ihdr_len + ihdr + w + h + rest)


class TestCocoKeypointsEval(unittest.TestCase):
    def test_build_gt_and_convert_predictions(self) -> None:
        repo_root = Path(__file__).resolve().parents[1]

        from yolozu.coco_keypoints_eval import build_coco_keypoints_ground_truth, predictions_to_coco_keypoints

        with tempfile.TemporaryDirectory(dir=str(repo_root)) as td:
            root = Path(td)
            img_path = root / "0001.png"
            _write_stub_png(img_path, width=64, height=32)

            records = [
                {
                    "image": str(img_path),
                    "labels": [
                        {
                            "class_id": 0,
                            "cx": 0.5,
                            "cy": 0.5,
                            "w": 0.4,
                            "h": 0.4,
                            "keypoints": [0.50, 0.50, 2, 0.60, 0.50, 1],
                        }
                    ],
                }
            ]

            gt, index = build_coco_keypoints_ground_truth(records, keypoints_format="xy_norm")
            self.assertEqual(int(index.keypoints_count), 2)
            self.assertEqual(len(gt.get("images") or []), 1)
            self.assertEqual(len(gt.get("annotations") or []), 1)

            ann = (gt.get("annotations") or [])[0]
            self.assertEqual(len(ann.get("keypoints") or []), 6)
            # kp0 x=0.50*64=32, y=0.50*32=16
            self.assertAlmostEqual(float(ann["keypoints"][0]), 32.0, places=6)
            self.assertAlmostEqual(float(ann["keypoints"][1]), 16.0, places=6)
            self.assertEqual(int(ann["keypoints"][2]), 2)
            self.assertEqual(int(ann["num_keypoints"]), 2)

            image_sizes = {img["id"]: (int(img["width"]), int(img["height"])) for img in gt.get("images") or []}
            preds = [
                {
                    "image": str(img_path),
                    "detections": [
                        {
                            "class_id": 0,
                            "score": 0.9,
                            "bbox": {"cx": 0.5, "cy": 0.5, "w": 0.4, "h": 0.4},
                            "keypoints": [0.50, 0.50, 1, 0.60, 0.50, 0.2],
                        }
                    ],
                }
            ]
            dt = predictions_to_coco_keypoints(
                preds,
                coco_index=index,
                image_sizes=image_sizes,
                keypoints_format="xy_norm",
            )
            self.assertEqual(len(dt), 1)
            self.assertEqual(len(dt[0].get("keypoints") or []), 6)
            self.assertAlmostEqual(float(dt[0]["keypoints"][5]), 0.2, places=6)

    def test_predictions_windows_style_image_key(self) -> None:
        repo_root = Path(__file__).resolve().parents[1]
        from yolozu.coco_keypoints_eval import build_coco_keypoints_ground_truth, predictions_to_coco_keypoints

        with tempfile.TemporaryDirectory(dir=str(repo_root)) as td:
            root = Path(td)
            img_path = root / "0001.png"
            _write_stub_png(img_path, width=64, height=32)

            records = [
                {
                    "image": str(img_path),
                    "labels": [
                        {
                            "class_id": 0,
                            "cx": 0.5,
                            "cy": 0.5,
                            "w": 0.4,
                            "h": 0.4,
                            "keypoints": [0.50, 0.50, 2, 0.60, 0.50, 1],
                        }
                    ],
                }
            ]

            gt, index = build_coco_keypoints_ground_truth(records, keypoints_format="xy_norm")
            image_sizes = {img["id"]: (int(img["width"]), int(img["height"])) for img in gt.get("images") or []}
            preds = [
                {
                    "image": str(img_path).replace("/", "\\"),
                    "detections": [
                        {
                            "class_id": 0,
                            "score": 0.9,
                            "bbox": {"cx": 0.5, "cy": 0.5, "w": 0.4, "h": 0.4},
                            "keypoints": [0.50, 0.50, 1, 0.60, 0.50, 0.2],
                        }
                    ],
                }
            ]
            dt = predictions_to_coco_keypoints(
                preds,
                coco_index=index,
                image_sizes=image_sizes,
                keypoints_format="xy_norm",
            )
            self.assertEqual(len(dt), 1)

    def test_oks_requires_pycocotools(self) -> None:
        from yolozu.coco_keypoints_eval import evaluate_coco_oks_map

        with self.assertRaises(RuntimeError):
            evaluate_coco_oks_map({"images": [], "annotations": [], "categories": []}, [])


if __name__ == "__main__":
    unittest.main()



from .constraints import apply_constraints
from .geometry import corrected_intrinsics, recover_translation


def evaluate_candidate(
    cfg,
    class_key,
    bbox_center,
    bbox_wh,
    offsets,
    z_pred,
    size_wh,
    k,
    k_delta,
    r_mat,
):
    k_prime = corrected_intrinsics(k, k_delta)
    t_xyz = recover_translation(bbox_center, offsets, z_pred, k_prime)
    constraints = apply_constraints(
        cfg,
        class_key=class_key,
        bbox_wh=bbox_wh,
        size_wh=size_wh,
        intrinsics_fx_fy=(k_prime[0], k_prime[1]),
        t_xyz=t_xyz,
        r_mat=r_mat,
        z_pred=z_pred,
    )
    return {"k_prime": k_prime, "t_xyz": t_xyz, "constraints": constraints}



import tempfile
import unittest
from pathlib import Path

from yolozu.cli_args import (
    parse_image_size_arg,
    require_non_negative_float,
    require_non_negative_int,
    require_positive_int,
    resolve_input_path,
    resolve_output_path,
)


class TestCliArgs(unittest.TestCase):
    def test_parse_image_size_variants(self):
        self.assertEqual(parse_image_size_arg("640"), (640, 640))
        self.assertEqual(parse_image_size_arg("640,480"), (640, 480))
        self.assertEqual(parse_image_size_arg("640x480"), (640, 480))
        with self.assertRaises(ValueError):
            parse_image_size_arg("0")

    def test_numeric_validators(self):
        self.assertEqual(require_non_negative_int(0, flag_name="--x"), 0)
        self.assertEqual(require_positive_int(1, flag_name="--x"), 1)
        self.assertEqual(require_non_negative_float(0.0, flag_name="--x"), 0.0)
        with self.assertRaises(ValueError):
            require_non_negative_int(-1, flag_name="--x")
        with self.assertRaises(ValueError):
            require_positive_int(0, flag_name="--x")
        with self.assertRaises(ValueError):
            require_non_negative_float(-0.1, flag_name="--x")

    def test_resolve_input_path_prefers_config_dir_when_exists(self):
        with tempfile.TemporaryDirectory() as td:
            root = Path(td)
            cwd = root / "cwd"
            cfg_dir = root / "cfg"
            repo = root / "repo"
            cwd.mkdir()
            cfg_dir.mkdir()
            repo.mkdir()
            target = cfg_dir / "predictions.json"
            target.write_text("[]", encoding="utf-8")

            resolved = resolve_input_path(
                "predictions.json",
                cwd=cwd,
                repo_root=repo,
                config_dir=cfg_dir,
            )
            self.assertEqual(resolved, target)

    def test_resolve_input_path_falls_back_to_repo_root(self):
        with tempfile.TemporaryDirectory() as td:
            root = Path(td)
            cwd = root / "cwd"
            repo = root / "repo"
            cwd.mkdir()
            repo.mkdir()
            target = repo / "dataset"
            target.mkdir()

            resolved = resolve_input_path("dataset", cwd=cwd, repo_root=repo)
            self.assertEqual(resolved, target)

    def test_resolve_output_path_uses_cwd(self):
        with tempfile.TemporaryDirectory() as td:
            cwd = Path(td)
            out = resolve_output_path("reports/out.json", cwd=cwd)
            self.assertEqual(out, cwd / "reports" / "out.json")


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import json
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Iterable


@dataclass(frozen=True)
class ValidationResult:
    warnings: list[str]


def _as_list(value: Any) -> list[Any]:
    if value is None:
        return []
    if isinstance(value, list):
        return value
    return [value]


def normalize_segmentation_predictions_json(data: Any) -> list[dict[str, Any]]:
    """Normalize supported segmentation prediction JSON shapes into a list of entries.

    Supported:
      1) [{"id": "...", "mask": "pred.png"}, ...]
      2) {"predictions": [ ...same as 1... ], ...}
      3) {"sample_id": "pred.png", ...}  (id->mask mapping)
    """

    if isinstance(data, dict) and "predictions" in data:
        data = data["predictions"]

    if isinstance(data, list):
        out: list[dict[str, Any]] = []
        for entry in data:
            if isinstance(entry, dict):
                out.append(entry)
        return out

    if isinstance(data, dict):
        out: list[dict[str, Any]] = []
        for sample_id, value in data.items():
            if isinstance(value, str):
                out.append({"id": str(sample_id), "mask": value})
            elif isinstance(value, dict):
                mask = value.get("mask")
                if isinstance(mask, str):
                    entry = dict(value)
                    entry["id"] = str(sample_id)
                    out.append(entry)
        return out

    raise ValueError("Unsupported segmentation predictions JSON format")


def normalize_segmentation_predictions_payload(data: Any) -> tuple[list[dict[str, Any]], dict[str, Any] | None]:
    """Normalize segmentation predictions JSON while preserving optional wrapped meta.

    Returns (entries, meta).
    """

    meta: dict[str, Any] | None = None
    if isinstance(data, dict) and "predictions" in data:
        raw_meta = data.get("meta")
        if raw_meta is not None:
            if not isinstance(raw_meta, dict):
                raise ValueError("meta must be an object when present")
            meta = raw_meta
        data = data["predictions"]

    return normalize_segmentation_predictions_json(data), meta


def validate_segmentation_predictions_entries(
    entries: Iterable[dict[str, Any]],
    *,
    where: str = "predictions",
) -> ValidationResult:
    warnings: list[str] = []
    seen: set[str] = set()
    for idx, entry in enumerate(entries):
        if not isinstance(entry, dict):
            raise ValueError(f"{where}[{idx}]: entry must be an object")
        sample_id = entry.get("id")
        if not isinstance(sample_id, str) or not sample_id:
            raise ValueError(f"{where}[{idx}]: missing 'id'")
        mask = entry.get("mask")
        if not isinstance(mask, str) or not mask:
            raise ValueError(f"{where}[{idx}]: missing 'mask'")
        if sample_id in seen:
            warnings.append(f"{where}[{idx}]: duplicate id: {sample_id}")
        seen.add(sample_id)
    return ValidationResult(warnings=warnings)


def validate_segmentation_predictions_payload(payload: Any) -> ValidationResult:
    entries = normalize_segmentation_predictions_json(payload)
    return validate_segmentation_predictions_entries(entries)


def load_segmentation_predictions_entries(path: str | Path) -> tuple[list[dict[str, Any]], dict[str, Any] | None]:
    path = Path(path)
    data = json.loads(path.read_text(encoding="utf-8"))
    entries, meta = normalize_segmentation_predictions_payload(data)
    validate_segmentation_predictions_entries(entries)
    return entries, meta


def build_id_to_mask(entries: Iterable[dict[str, Any]]) -> dict[str, str]:
    out: dict[str, str] = {}
    for entry in entries:
        sample_id = entry.get("id")
        mask = entry.get("mask")
        if isinstance(sample_id, str) and isinstance(mask, str):
            out[sample_id] = mask
    return out




import unittest

from yolozu.simple_map import evaluate_map


class TestSimpleMap(unittest.TestCase):
    def test_distillation_improves_map(self):
        records = [
            {
                "image": "img1.jpg",
                "labels": [
                    {"class_id": 0, "cx": 0.5, "cy": 0.5, "w": 0.2, "h": 0.2},
                ],
            },
            {
                "image": "img2.jpg",
                "labels": [
                    {"class_id": 0, "cx": 0.3, "cy": 0.3, "w": 0.2, "h": 0.2},
                ],
            },
        ]

        student = [
            {
                "image": "img1.jpg",
                "detections": [
                    {"class_id": 0, "score": 0.4, "bbox": {"cx": 0.5, "cy": 0.5, "w": 0.2, "h": 0.2}},
                ],
            },
            {"image": "img2.jpg", "detections": []},
        ]

        teacher = [
            {
                "image": "img1.jpg",
                "detections": [
                    {"class_id": 0, "score": 0.9, "bbox": {"cx": 0.5, "cy": 0.5, "w": 0.2, "h": 0.2}},
                ],
            },
            {
                "image": "img2.jpg",
                "detections": [
                    {"class_id": 0, "score": 0.9, "bbox": {"cx": 0.3, "cy": 0.3, "w": 0.2, "h": 0.2}},
                ],
            },
        ]

        student_map = evaluate_map(records, student, iou_thresholds=[0.5]).map50
        teacher_map = evaluate_map(records, teacher, iou_thresholds=[0.5]).map50
        self.assertGreaterEqual(teacher_map, student_map)

    def test_map_matches_windows_style_prediction_image(self):
        records = [
            {
                "image": "/dataset/images/img1.jpg",
                "labels": [{"class_id": 0, "cx": 0.5, "cy": 0.5, "w": 0.2, "h": 0.2}],
            }
        ]
        preds = [
            {
                "image": r"C:\dataset\images\img1.jpg",
                "detections": [{"class_id": 0, "score": 0.9, "bbox": {"cx": 0.5, "cy": 0.5, "w": 0.2, "h": 0.2}}],
            }
        ]
        result = evaluate_map(records, preds, iou_thresholds=[0.5])
        self.assertAlmostEqual(float(result.map50), 1.0, places=6)


if __name__ == "__main__":
    unittest.main()



import sys
import unittest
from pathlib import Path

sys.path.insert(0, str(Path(__file__).resolve().parents[1]))

try:
    import torch
    from torch import nn
except ImportError:  # pragma: no cover
    torch = None
    nn = None

from yolozu.tta.ttt_mim import (
    apply_mask,
    filter_parameters,
    generate_block_mask,
    run_ttt_mim,
)


@unittest.skipIf(torch is None, "torch not installed")
class TestTTTMIM(unittest.TestCase):
    def test_generate_block_mask_full(self):
        mask = generate_block_mask(4, 4, patch_size=2, mask_prob=1.0)
        self.assertTrue(mask.all())

    def test_apply_mask(self):
        x = torch.ones(1, 3, 4, 4)
        mask = torch.zeros(4, 4, dtype=torch.bool)
        mask[0:2, 0:2] = True
        masked = apply_mask(x, mask, mask_value=0.0)
        self.assertTrue(torch.allclose(masked[0, :, 0, 0], torch.zeros(3)))
        self.assertTrue(torch.allclose(masked[0, :, 3, 3], torch.ones(3)))

    def test_filter_parameters_include_exclude(self):
        model = nn.Sequential(nn.Conv2d(3, 4, 1), nn.Conv2d(4, 2, 1))
        params = filter_parameters(model.named_parameters(), include=["0"], exclude=["bias"])
        self.assertTrue(params)
        for p in params:
            self.assertTrue(p.requires_grad)

    def test_run_ttt_mim(self):
        torch.manual_seed(0)
        model = nn.Conv2d(3, 3, kernel_size=1, bias=False)
        x = torch.rand(1, 3, 8, 8)
        res = run_ttt_mim(model, x, steps=2, mask_prob=0.5, patch_size=2)
        self.assertEqual(len(res.losses), 2)
        self.assertGreaterEqual(res.mask_ratio, 0.0)
        self.assertLessEqual(res.mask_ratio, 1.0)
        self.assertGreater(res.updated_param_count, 0)

    def test_update_filter_norm_only(self):
        model = nn.Sequential(nn.Conv2d(3, 3, 1), nn.BatchNorm2d(3))
        x = torch.rand(1, 3, 4, 4)
        res = run_ttt_mim(model, x, steps=1, mask_prob=0.5, patch_size=2, update_filter="norm_only")
        expected = sum(p.numel() for p in model[1].parameters())
        self.assertEqual(res.updated_param_count, expected)

    def test_update_filter_adapter_only(self):
        class Model(nn.Module):
            def __init__(self):
                super().__init__()
                self.adapter = nn.Conv2d(3, 3, 1)
                self.head = nn.Conv2d(3, 3, 1)

            def forward(self, x):
                return self.adapter(x)

        model = Model()
        x = torch.rand(1, 3, 4, 4)
        res = run_ttt_mim(model, x, steps=1, mask_prob=0.5, patch_size=2, update_filter="adapter_only")
        expected = sum(p.numel() for p in model.adapter.parameters())
        self.assertEqual(res.updated_param_count, expected)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from .image_keys import add_image_aliases, lookup_image_alias
from .simple_map import _bbox_iou_cxcywh_norm


@dataclass(frozen=True)
class DistillStats:
    matched: int
    added: int
    avg_score_gap: float


def _index_by_image(entries: list[dict[str, Any]]) -> dict[str, list[dict[str, Any]]]:
    index: dict[str, list[dict[str, Any]]] = {}
    for entry in entries:
        image = str(entry.get("image", ""))
        if not image:
            continue
        dets = entry.get("detections", []) or []
        add_image_aliases(index, image, list(dets) if isinstance(dets, list) else [])
    return index


def distill_predictions(
    student_entries: list[dict[str, Any]],
    teacher_entries: list[dict[str, Any]],
    *,
    iou_threshold: float = 0.7,
    alpha: float = 0.5,
    add_missing: bool = True,
    add_score_scale: float = 0.5,
) -> tuple[list[dict[str, Any]], DistillStats]:
    if not 0.0 <= alpha <= 1.0:
        raise ValueError("alpha must be in [0,1]")
    if iou_threshold < 0.0:
        raise ValueError("iou_threshold must be >= 0")

    teacher_index = _index_by_image(teacher_entries)

    out_entries: list[dict[str, Any]] = []
    matched = 0
    added = 0
    total_gap = 0.0

    for entry in student_entries:
        image = str(entry.get("image", ""))
        if not image:
            continue
        student_dets = [dict(d) for d in (entry.get("detections", []) or [])]
        teacher_dets = lookup_image_alias(teacher_index, image) or []
        teacher_used = [False] * len(teacher_dets)

        for det in student_dets:
            best_iou = 0.0
            best_idx = -1
            for idx, tdet in enumerate(teacher_dets):
                if int(tdet.get("class_id", -1)) != int(det.get("class_id", -2)):
                    continue
                iou = _bbox_iou_cxcywh_norm(det.get("bbox", {}), tdet.get("bbox", {}))
                if iou > best_iou:
                    best_iou = iou
                    best_idx = idx
            if best_iou >= iou_threshold and best_idx >= 0:
                teacher_used[best_idx] = True
                t_score = float(teacher_dets[best_idx].get("score", 0.0))
                s_score = float(det.get("score", 0.0))
                total_gap += abs(t_score - s_score)
                matched += 1
                det["score"] = max(s_score, alpha * t_score + (1.0 - alpha) * s_score)

        if add_missing:
            for idx, tdet in enumerate(teacher_dets):
                if teacher_used[idx]:
                    continue
                score = float(tdet.get("score", 0.0)) * float(add_score_scale)
                det_out = dict(tdet)
                det_out["score"] = score
                student_dets.append(det_out)
                added += 1

        out_entries.append({"image": image, "detections": student_dets})

    avg_gap = total_gap / float(max(1, matched))
    return out_entries, DistillStats(matched=matched, added=added, avg_score_gap=float(avg_gap))



from __future__ import annotations

import argparse
import json
import sys
from pathlib import Path

from yolozu import __version__

from .cli_args import parse_image_size_arg, require_non_negative_int
from .config import simple_yaml_load


def _load_config(path: Path) -> dict:
    text = path.read_text(encoding="utf-8")
    if path.suffix.lower() in (".yaml", ".yml"):
        try:
            import yaml  # type: ignore

            data = yaml.safe_load(text)
            return data or {}
        except Exception:
            return simple_yaml_load(text)
    if path.suffix.lower() == ".json":
        return json.loads(text)
    try:
        return json.loads(text)
    except Exception:
        return simple_yaml_load(text)


def _build_args_from_config(cfg: dict) -> list[str]:
    args: list[str] = []
    for key, value in cfg.items():
        if value is None:
            continue
        arg = f"--{str(key).replace('_', '-') }"
        if isinstance(value, bool):
            if value:
                args.append(arg)
            continue
        if isinstance(value, (list, tuple)):
            args.append(arg)
            args.extend([str(v) for v in value])
            continue
        args.append(arg)
        args.append(str(value))
    return args


def _cmd_train(config_path: Path) -> int:
    try:
        from rtdetr_pose.tools.train_minimal import main as train_main
    except Exception as exc:  # pragma: no cover
        raise SystemExit(
            "yolozu train is currently supported from a source checkout (e.g. `pip install -e .`) "
            "because it depends on in-repo trainer scaffolding under rtdetr_pose/tools."
        ) from exc

    return int(train_main(["--config", str(config_path)]))


def _cmd_test(config_path: Path) -> int:
    try:
        from tools.run_scenarios import main as scenarios_main
    except Exception as exc:  # pragma: no cover
        raise SystemExit(
            "yolozu test is currently supported from a source checkout (e.g. `pip install -e .`) "
            "because it depends on in-repo tools/ scripts."
        ) from exc

    cfg = _load_config(config_path)
    args = _build_args_from_config(cfg)
    scenarios_main(args)
    return 0


def _cmd_doctor(output: str) -> int:
    from yolozu.doctor import write_doctor_report

    return int(write_doctor_report(output=output))


def _cmd_export(args: argparse.Namespace) -> int:
    from yolozu.export import (
        DEFAULT_PREDICTIONS_PATH,
        export_dummy_predictions,
        export_labels_predictions,
        write_predictions_json,
    )

    backend = str(getattr(args, "backend", "dummy"))

    dataset = str(args.dataset)
    if not dataset:
        raise SystemExit("--dataset is required")

    try:
        if backend == "dummy":
            payload, _run = export_dummy_predictions(
                dataset_root=dataset,
                split=str(args.split) if args.split else None,
                max_images=int(args.max_images) if args.max_images is not None else None,
                score=float(args.score),
            )
        elif backend == "labels":
            payload, _run = export_labels_predictions(
                dataset_root=dataset,
                split=str(args.split) if args.split else None,
                max_images=int(args.max_images) if args.max_images is not None else None,
                score=float(args.score),
            )
        else:
            raise SystemExit(f"unsupported --backend: {backend}")
    except FileNotFoundError as exc:
        raise SystemExit(str(exc)) from exc

    output = str(args.output) if args.output else DEFAULT_PREDICTIONS_PATH
    out_path = write_predictions_json(output=output, payload=payload, force=bool(args.force))
    print(str(out_path))
    return 0


def _cmd_validate(args: argparse.Namespace) -> int:
    if args.validate_command == "dataset":
        from yolozu.dataset import build_manifest
        from yolozu.dataset_validator import validate_dataset_records

        manifest = build_manifest(str(args.dataset), split=str(args.split) if args.split else None)
        records = manifest.get("images") or []
        if not isinstance(records, list):
            raise SystemExit("invalid dataset manifest (expected list under 'images')")
        if args.max_images is not None:
            records = records[: int(args.max_images)]

        res = validate_dataset_records(
            records,
            strict=bool(args.strict),
            mode=str(args.mode),
            check_images=not bool(args.no_check_images),
        )
        for w in res.warnings:
            print(w, file=sys.stderr)
        if res.errors:
            for e in res.errors:
                print(e, file=sys.stderr)
            return 1
        return 0

    path = Path(str(args.path))
    if not path.exists():
        raise SystemExit(f"file not found: {path}")
    try:
        payload = json.loads(path.read_text(encoding="utf-8"))
    except Exception as exc:
        raise SystemExit(f"failed to parse json: {path} ({exc})") from exc

    if args.validate_command == "predictions":
        from yolozu.predictions import validate_predictions_payload

        try:
            res = validate_predictions_payload(payload, strict=bool(args.strict))
        except Exception as exc:
            raise SystemExit(str(exc)) from exc
        for w in res.warnings:
            print(w, file=sys.stderr)
        return 0

    if args.validate_command == "instance-seg":
        from yolozu.instance_segmentation_predictions import (
            normalize_instance_segmentation_predictions_payload,
            validate_instance_segmentation_predictions_entries,
        )

        try:
            entries, _meta = normalize_instance_segmentation_predictions_payload(payload)
            res = validate_instance_segmentation_predictions_entries(entries, where="predictions")
        except Exception as exc:
            raise SystemExit(str(exc)) from exc
        for w in res.warnings:
            print(w, file=sys.stderr)
        return 0

    raise SystemExit("unknown validate command")


def _cmd_eval_instance_seg(args: argparse.Namespace) -> int:
    from yolozu.instance_segmentation_report import run_instance_segmentation_eval

    out_json, out_html = run_instance_segmentation_eval(
        dataset_root=str(args.dataset),
        split=str(args.split) if args.split else None,
        predictions=str(args.predictions),
        pred_root=str(args.pred_root) if args.pred_root else None,
        classes=str(args.classes) if args.classes else None,
        output=str(args.output),
        html=str(args.html) if args.html else None,
        title=str(args.title),
        overlays_dir=str(args.overlays_dir) if args.overlays_dir else None,
        max_overlays=int(args.max_overlays),
        overlay_sort=str(args.overlay_sort),
        overlay_max_size=int(args.overlay_max_size),
        overlay_alpha=float(args.overlay_alpha),
        min_score=float(args.min_score),
        max_images=int(args.max_images) if args.max_images is not None else None,
        diag_iou=float(args.diag_iou),
        per_image_limit=int(args.per_image_limit),
        allow_rgb_masks=bool(args.allow_rgb_masks),
    )
    print(str(out_json))
    if out_html is not None:
        print(str(out_html))
    return 0


def _cmd_onnxrt_export(args: argparse.Namespace) -> int:
    from yolozu.onnxrt_export import export_predictions_onnxrt, write_predictions_json

    try:
        payload = export_predictions_onnxrt(
            dataset_root=str(args.dataset),
            split=str(args.split) if args.split else None,
            max_images=int(args.max_images) if args.max_images is not None else None,
            onnx=str(args.onnx) if args.onnx else None,
            input_name=str(args.input_name),
            boxes_output=str(args.boxes_output),
            scores_output=str(args.scores_output),
            class_output=(str(args.class_output) if args.class_output else None),
            combined_output=(str(args.combined_output) if args.combined_output else None),
            combined_format=str(args.combined_format),
            raw_output=(str(args.raw_output) if args.raw_output else None),
            raw_format=str(args.raw_format),
            raw_postprocess=str(args.raw_postprocess),
            boxes_format=str(args.boxes_format),
            boxes_scale=str(args.boxes_scale),
            min_score=float(args.min_score),
            topk=int(args.topk),
            nms_iou=float(args.nms_iou),
            agnostic_nms=bool(args.agnostic_nms),
            imgsz=int(args.imgsz),
            dry_run=bool(args.dry_run),
            strict=bool(args.strict),
        )
    except Exception as exc:
        raise SystemExit(str(exc)) from exc

    out_path = write_predictions_json(output=str(args.output), payload=payload, force=bool(args.force))
    print(str(out_path))
    return 0


def _cmd_predict_images(args: argparse.Namespace) -> int:
    from yolozu.predict_images import predict_images

    try:
        max_images = require_non_negative_int(args.max_images, flag_name="--max-images")
    except ValueError as exc:
        raise SystemExit(str(exc)) from exc

    try:
        out_json, out_html = predict_images(
            backend=str(args.backend),
            input_dir=str(args.input_dir),
            output=str(args.output),
            score=float(args.score),
            max_images=max_images,
            force=bool(args.force),
            glob_patterns=list(args.glob) if args.glob else None,
            overlays_dir=str(args.overlays_dir) if args.overlays_dir else None,
            html=str(args.html) if args.html else None,
            title=str(args.title),
            onnx=(str(args.onnx) if args.onnx else None),
            input_name=str(args.input_name),
            boxes_output=str(args.boxes_output),
            scores_output=str(args.scores_output),
            class_output=(str(args.class_output) if args.class_output else None),
            combined_output=(str(args.combined_output) if args.combined_output else None),
            combined_format=str(args.combined_format),
            raw_output=(str(args.raw_output) if args.raw_output else None),
            raw_format=str(args.raw_format),
            raw_postprocess=str(args.raw_postprocess),
            boxes_format=str(args.boxes_format),
            boxes_scale=str(args.boxes_scale),
            min_score=float(args.min_score),
            topk=int(args.topk),
            nms_iou=float(args.nms_iou),
            agnostic_nms=bool(args.agnostic_nms),
            imgsz=int(args.imgsz),
            dry_run=bool(args.dry_run),
            strict=bool(args.strict),
        )
    except Exception as exc:
        raise SystemExit(str(exc)) from exc

    print(str(out_json))
    if out_html is not None:
        print(str(out_html))
    return 0


def _cmd_eval_coco(args: argparse.Namespace) -> int:
    import time

    from yolozu.coco_eval import build_coco_ground_truth, evaluate_coco_map, predictions_to_coco_detections
    from yolozu.dataset import build_manifest
    from yolozu.predictions import load_predictions_entries, validate_predictions_entries

    dataset_root = Path(str(args.dataset)).expanduser()
    if not dataset_root.is_absolute():
        dataset_root = Path.cwd() / dataset_root

    manifest = build_manifest(dataset_root, split=str(args.split) if args.split else None)
    records = list(manifest.get("images") or [])
    if args.max_images is not None:
        records = records[: int(args.max_images)]

    gt, index = build_coco_ground_truth(records)
    image_sizes = {image["id"]: (int(image["width"]), int(image["height"])) for image in gt["images"]}

    predictions_path = Path(str(args.predictions)).expanduser()
    if not predictions_path.is_absolute():
        predictions_path = Path.cwd() / predictions_path
    predictions = load_predictions_entries(predictions_path)
    validation = validate_predictions_entries(predictions, strict=False)
    detections = predictions_to_coco_detections(
        predictions,
        coco_index=index,
        image_sizes=image_sizes,
        bbox_format=str(args.bbox_format),
    )

    if bool(args.dry_run):
        result: dict[str, object] = {
            "metrics": {"map50_95": None, "map50": None, "map75": None, "ar100": None},
            "stats": [],
            "dry_run": True,
            "counts": {"images": int(len(records)), "detections": int(len(detections))},
            "warnings": validation.warnings,
        }
    else:
        result = evaluate_coco_map(gt, detections)
        result["warnings"] = validation.warnings

    payload: dict[str, object] = {
        "report_schema_version": 1,
        "timestamp": time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime()),
        "dataset": str(dataset_root),
        "split": manifest.get("split"),
        "split_requested": str(args.split) if args.split else None,
        "predictions": str(predictions_path),
        "bbox_format": str(args.bbox_format),
        "max_images": int(args.max_images) if args.max_images is not None else None,
        **result,
    }

    output_path = Path(str(args.output)).expanduser()
    if not output_path.is_absolute():
        output_path = Path.cwd() / output_path
    output_path.parent.mkdir(parents=True, exist_ok=True)
    output_path.write_text(json.dumps(payload, indent=2, sort_keys=True, ensure_ascii=False) + "\n", encoding="utf-8")
    print(str(output_path))
    return 0


def _cmd_parity(args: argparse.Namespace) -> int:
    from yolozu.predictions_parity import compare_predictions

    try:
        max_images = require_non_negative_int(args.max_images, flag_name="--max-images")
        image_size = parse_image_size_arg(args.image_size, flag_name="--image-size")
    except ValueError as exc:
        raise SystemExit(str(exc)) from exc

    report = compare_predictions(
        reference=str(args.reference),
        candidate=str(args.candidate),
        image_size=image_size,
        max_images=max_images,
        iou_thresh=float(args.iou_thresh),
        score_atol=float(args.score_atol),
        bbox_atol=float(args.bbox_atol),
    )
    print(json.dumps(report, indent=2, sort_keys=True, ensure_ascii=False))
    return 0 if bool(report.get("ok")) else 2


def _cmd_resources(args: argparse.Namespace) -> int:
    from yolozu import resources

    if args.resources_command == "list":
        for p in resources.list_resource_paths():
            print(p)
        return 0

    if args.resources_command == "cat":
        text = resources.read_text(str(args.path))
        print(text, end="" if text.endswith("\n") else "\n")
        return 0

    if args.resources_command == "copy":
        out = resources.copy_to(str(args.path), output=str(args.output), force=bool(args.force))
        print(str(out))
        return 0

    raise SystemExit("unknown resources command")


def main(argv: list[str] | None = None) -> int:
    parser = argparse.ArgumentParser(prog="yolozu")
    parser.add_argument("--version", action="version", version=f"%(prog)s {__version__}")
    sub = parser.add_subparsers(dest="command", required=True)

    doctor = sub.add_parser("doctor", help="Print environment diagnostics as JSON.")
    doctor.add_argument("--output", default="reports/doctor.json", help="Output JSON path (use - for stdout).")

    export = sub.add_parser("export", help="Export predictions.json artifacts.")
    export.add_argument(
        "--backend",
        choices=("dummy", "labels"),
        default="dummy",
        help="Export backend (dummy=1 det/image; labels=use dataset labels).",
    )
    export.add_argument("--dataset", default="data/coco128", help="YOLO-format dataset root.")
    export.add_argument("--split", default=None, help="Dataset split under images/ and labels/ (default: auto).")
    export.add_argument("--max-images", type=int, default=50, help="Optional cap for number of images.")
    export.add_argument("--score", type=float, default=0.9, help="Score to assign to exported detections (default: 0.9).")
    export.add_argument(
        "--output",
        default=None,
        help="Predictions JSON output path (default: reports/predictions.json).",
    )
    export.add_argument("--force", action="store_true", help="Overwrite outputs if they exist.")

    predict = sub.add_parser("predict-images", help="Run folder inference and write predictions JSON + overlays + HTML.")
    predict.add_argument("--backend", choices=("dummy", "onnxrt"), default="dummy", help="Inference backend.")
    predict.add_argument("--input-dir", required=True, help="Input directory containing images.")
    predict.add_argument("--glob", action="append", default=None, help="Glob pattern(s) under input dir (repeatable).")
    predict.add_argument("--max-images", type=int, default=None, help="Optional cap for number of images.")
    predict.add_argument("--score", type=float, default=0.9, help="Dummy score when --backend=dummy.")
    predict.add_argument("--output", default="reports/predict_images.json", help="Predictions JSON output path.")
    predict.add_argument("--force", action="store_true", help="Overwrite outputs if they exist.")
    predict.add_argument("--overlays-dir", default="reports/overlays", help="Overlay images output directory.")
    predict.add_argument("--html", default="reports/predict_images.html", help="Optional HTML report path.")
    predict.add_argument("--title", default="YOLOZU predict-images report", help="HTML title.")
    predict.add_argument("--onnx", default=None, help="Path to ONNX model (required for --backend onnxrt unless --dry-run).")
    predict.add_argument("--input-name", default="images", help="ONNX input name (default: images).")
    predict.add_argument("--boxes-output", default="boxes", help="Output name for boxes tensor (default: boxes).")
    predict.add_argument("--scores-output", default="scores", help="Output name for scores tensor (default: scores).")
    predict.add_argument("--class-output", default=None, help="Optional output name for class_id tensor.")
    predict.add_argument("--combined-output", default=None, help="Optional output name for [x1,y1,x2,y2,score,class_id].")
    predict.add_argument("--combined-format", choices=("xyxy_score_class",), default="xyxy_score_class")
    predict.add_argument("--raw-output", default=None, help="Optional output name for raw head output.")
    predict.add_argument("--raw-format", choices=("yolo_84",), default="yolo_84")
    predict.add_argument("--raw-postprocess", choices=("native", "ultralytics"), default="native")
    predict.add_argument("--boxes-format", choices=("xyxy",), default="xyxy")
    predict.add_argument("--boxes-scale", choices=("abs", "norm"), default="norm")
    predict.add_argument("--min-score", type=float, default=0.001, help="Score threshold (default: 0.001).")
    predict.add_argument("--topk", type=int, default=300, help="Top-K detections per image (default: 300).")
    predict.add_argument("--nms-iou", type=float, default=0.7, help="NMS IoU for raw output decode (default: 0.7).")
    predict.add_argument("--agnostic-nms", action="store_true", help="Class-agnostic NMS for raw output decode.")
    predict.add_argument("--imgsz", type=int, default=640, help="Input image size (square, default: 640).")
    predict.add_argument("--dry-run", action="store_true", help="Write schema-correct JSON without running inference.")
    predict.add_argument("--strict", action="store_true", help="Strict prediction schema validation before writing.")

    eval_coco = sub.add_parser("eval-coco", help="Evaluate detections with COCOeval (optional extra: yolozu[coco]).")
    eval_coco.add_argument("--dataset", required=True, help="YOLO-format dataset root (images/ + labels/).")
    eval_coco.add_argument("--split", default=None, help="Dataset split under images/ and labels/ (default: auto).")
    eval_coco.add_argument("--predictions", required=True, help="Predictions JSON path.")
    eval_coco.add_argument(
        "--bbox-format",
        choices=("cxcywh_norm", "cxcywh_abs", "xywh_abs", "xyxy_abs"),
        default="cxcywh_norm",
        help="How to interpret detection bbox fields (default: cxcywh_norm).",
    )
    eval_coco.add_argument("--dry-run", action="store_true", help="Skip COCOeval; only validate/convert predictions.")
    eval_coco.add_argument("--max-images", type=int, default=None, help="Optional cap for number of images.")
    eval_coco.add_argument("--output", default="reports/coco_eval.json", help="Output report path.")

    parity = sub.add_parser("parity", help="Compare two predictions JSON artifacts for backend parity.")
    parity.add_argument("--reference", required=True, help="Reference predictions JSON (e.g. PyTorch).")
    parity.add_argument("--candidate", required=True, help="Candidate predictions JSON (e.g. ONNXRuntime).")
    parity.add_argument("--iou-thresh", type=float, default=0.99, help="IoU threshold for a match.")
    parity.add_argument("--score-atol", type=float, default=1e-4, help="Absolute tolerance for score differences.")
    parity.add_argument("--bbox-atol", type=float, default=1e-4, help="Absolute tolerance for bbox differences.")
    parity.add_argument("--max-images", type=int, default=None, help="Optional cap for number of images.")
    parity.add_argument("--image-size", default=None, help="Optional fixed image size (N or W,H) to skip image reads.")

    validate = sub.add_parser("validate", help="Validate artifacts (predictions JSON, instance-seg predictions).")
    validate_sub = validate.add_subparsers(dest="validate_command", required=True)
    val_pred = validate_sub.add_parser("predictions", help="Validate predictions JSON (detections+bbox schema).")
    val_pred.add_argument("path", type=str, help="Path to predictions JSON (list or wrapper).")
    val_pred.add_argument("--strict", action="store_true", help="Strict validation (types, required keys).")

    val_is = validate_sub.add_parser("instance-seg", help="Validate instance-segmentation predictions JSON (PNG masks).")
    val_is.add_argument("path", type=str, help="Path to instance-seg predictions JSON.")

    val_ds = validate_sub.add_parser("dataset", help="Validate a YOLO-format dataset layout + labels.")
    val_ds.add_argument("dataset", type=str, help="YOLO-format dataset root (contains images/ + labels/).")
    val_ds.add_argument("--split", default=None, help="Split under images/ and labels/ (default: auto).")
    val_ds.add_argument("--max-images", type=int, default=None, help="Optional cap for number of images checked.")
    val_ds.add_argument("--strict", action="store_true", help="Strict bbox checks (range + inside-image).")
    val_ds.add_argument("--mode", choices=("fail", "warn"), default="fail", help="fail=exit nonzero on errors; warn=always exit 0.")
    val_ds.add_argument("--no-check-images", action="store_true", help="Skip image existence/size checks.")

    eis = sub.add_parser(
        "eval-instance-seg",
        help="Evaluate instance segmentation predictions (mask mAP over PNG masks).",
    )
    eis.add_argument("--dataset", required=True, help="YOLO-format dataset root (images/ + labels/).")
    eis.add_argument("--split", default=None, help="Split under images/ and labels/ (default: auto).")
    eis.add_argument("--predictions", required=True, help="Instance segmentation predictions JSON.")
    eis.add_argument("--pred-root", default=None, help="Optional root to resolve relative prediction mask paths.")
    eis.add_argument("--classes", default=None, help="Optional classes.txt/classes.json for class_id→name.")
    eis.add_argument("--output", default="reports/instance_seg_eval.json", help="Output JSON report path.")
    eis.add_argument("--html", default=None, help="Optional HTML report path.")
    eis.add_argument("--title", default="YOLOZU instance segmentation eval report", help="HTML title.")
    eis.add_argument("--overlays-dir", default=None, help="Optional directory to write overlay images for HTML.")
    eis.add_argument("--max-overlays", type=int, default=0, help="Max overlays to render (default: 0).")
    eis.add_argument(
        "--overlay-sort",
        choices=("worst", "best", "first"),
        default="worst",
        help="How to select overlay samples (default: worst).",
    )
    eis.add_argument("--overlay-max-size", type=int, default=768, help="Max size (max(H,W)) for overlay images (default: 768).")
    eis.add_argument("--overlay-alpha", type=float, default=0.5, help="Mask overlay alpha (default: 0.5).")
    eis.add_argument("--min-score", type=float, default=0.0, help="Minimum score threshold for predictions (default: 0.0).")
    eis.add_argument("--max-images", type=int, default=None, help="Optional cap for number of images to evaluate.")
    eis.add_argument("--diag-iou", type=float, default=0.5, help="IoU threshold used for per-image diagnostics/overlay selection (default: 0.5).")
    eis.add_argument("--per-image-limit", type=int, default=100, help="How many per-image rows to store in the report/meta and HTML (default: 100).")
    eis.add_argument(
        "--allow-rgb-masks",
        action="store_true",
        help="Allow 3-channel masks (uses channel 0; intended for grayscale stored as RGB).",
    )

    onnxrt = sub.add_parser("onnxrt", help="ONNXRuntime utilities (optional extra: yolozu[onnxrt]).")
    onnxrt_sub = onnxrt.add_subparsers(dest="onnxrt_command", required=True)
    onnxrt_export = onnxrt_sub.add_parser("export", help="Run ONNXRuntime inference and export predictions JSON.")
    onnxrt_export.add_argument("--dataset", required=True, help="YOLO-format dataset root (images/ + labels/).")
    onnxrt_export.add_argument("--split", default=None, help="Split under images/ and labels/ (default: auto).")
    onnxrt_export.add_argument("--max-images", type=int, default=None, help="Optional cap for quick runs.")
    onnxrt_export.add_argument("--onnx", default=None, help="Path to ONNX model (required unless --dry-run).")
    onnxrt_export.add_argument("--input-name", default="images", help="ONNX input name (default: images).")
    onnxrt_export.add_argument("--boxes-output", default="boxes", help="Output name for boxes tensor (default: boxes).")
    onnxrt_export.add_argument("--scores-output", default="scores", help="Output name for scores tensor (default: scores).")
    onnxrt_export.add_argument("--class-output", default=None, help="Optional output name for class_id tensor (default: none).")
    onnxrt_export.add_argument(
        "--combined-output",
        default=None,
        help="Optional single output name with (N,6) or (1,N,6) entries [x1,y1,x2,y2,score,class_id].",
    )
    onnxrt_export.add_argument(
        "--combined-format",
        choices=("xyxy_score_class",),
        default="xyxy_score_class",
        help="Layout for --combined-output (default: xyxy_score_class).",
    )
    onnxrt_export.add_argument(
        "--raw-output",
        default=None,
        help="Optional single output name with raw head output (e.g., 1x84x8400) to decode + NMS.",
    )
    onnxrt_export.add_argument(
        "--raw-format",
        choices=("yolo_84",),
        default="yolo_84",
        help="Layout for --raw-output (default: yolo_84).",
    )
    onnxrt_export.add_argument(
        "--raw-postprocess",
        choices=("native", "ultralytics"),
        default="native",
        help="Postprocess for --raw-output (default: native).",
    )
    onnxrt_export.add_argument(
        "--boxes-format",
        choices=("xyxy",),
        default="xyxy",
        help="Box layout produced by the model in input-image space (default: xyxy).",
    )
    onnxrt_export.add_argument(
        "--boxes-scale",
        choices=("abs", "norm"),
        default="norm",
        help="Whether boxes are in pixels (abs) or normalized [0,1] wrt input_size (default: norm).",
    )
    onnxrt_export.add_argument("--min-score", type=float, default=0.001, help="Score threshold (no NMS).")
    onnxrt_export.add_argument("--topk", type=int, default=300, help="Keep top-K detections per image (no NMS).")
    onnxrt_export.add_argument("--nms-iou", type=float, default=0.7, help="IoU threshold for NMS when decoding raw output.")
    onnxrt_export.add_argument(
        "--agnostic-nms",
        action="store_true",
        help="Use class-agnostic NMS when decoding raw output.",
    )
    onnxrt_export.add_argument("--imgsz", type=int, default=640, help="Input image size (square, default: 640).")
    onnxrt_export.add_argument("--output", default="reports/predictions_onnxrt.json", help="Where to write predictions JSON.")
    onnxrt_export.add_argument("--force", action="store_true", help="Overwrite outputs if they exist.")
    onnxrt_export.add_argument("--dry-run", action="store_true", help="Write schema-correct JSON without running inference.")
    onnxrt_export.add_argument("--strict", action="store_true", help="Strict prediction schema validation before writing.")

    resources_p = sub.add_parser("resources", help="Access packaged configs/schemas/protocols.")
    resources_sub = resources_p.add_subparsers(dest="resources_command", required=True)
    resources_sub.add_parser("list", help="List packaged resource paths.")
    cat = resources_sub.add_parser("cat", help="Print a packaged resource to stdout.")
    cat.add_argument("path", type=str, help="Resource path under yolozu/data (e.g., schemas/predictions.schema.json).")
    copy = resources_sub.add_parser("copy", help="Copy a packaged resource to a file path.")
    copy.add_argument("path", type=str, help="Resource path under yolozu/data.")
    copy.add_argument("--output", required=True, help="Output file path.")
    copy.add_argument("--force", action="store_true", help="Overwrite output if it exists.")

    repo_root = Path(__file__).resolve().parents[1]
    dev_enabled = (repo_root / "rtdetr_pose" / "tools" / "train_minimal.py").exists() and (
        repo_root / "tools" / "run_scenarios.py"
    ).exists()
    dev_help = "(dev) Source-checkout-only commands (train/test)."
    if not dev_enabled:
        dev_help = argparse.SUPPRESS

    dev = sub.add_parser("dev", help=dev_help)
    dev_sub = dev.add_subparsers(dest="dev_command", required=True)

    dev_train = dev_sub.add_parser("train", help="Run training using a YAML/JSON config (source checkout only).")
    dev_train.add_argument("config", type=str, help="Path to train config (e.g. configs/examples/train_setting.yaml).")

    dev_test = dev_sub.add_parser("test", help="Run scenario tests using a YAML/JSON config (source checkout only).")
    dev_test.add_argument("config", type=str, help="Path to test config (e.g. configs/examples/test_setting.yaml).")

    # Backward-compatible hidden aliases.
    alias_help = argparse.SUPPRESS
    alias_train = sub.add_parser("train", help=alias_help)
    alias_train.add_argument("config", type=str, help="Path to train config (e.g. configs/examples/train_setting.yaml).")
    alias_test = sub.add_parser("test", help=alias_help)
    alias_test.add_argument("config", type=str, help="Path to test config (e.g. configs/examples/test_setting.yaml).")

    demo = sub.add_parser("demo", help="Run small self-contained demos (CPU-friendly).")
    demo_sub = demo.add_subparsers(dest="demo_command", required=True)

    demo_is = demo_sub.add_parser("instance-seg", help="Synthetic instance-seg eval demo (numpy + Pillow).")
    demo_is.add_argument("--run-dir", default=None, help="Run directory (default: runs/yolozu_demos/instance_seg/<utc>).")
    demo_is.add_argument("--seed", type=int, default=0, help="Random seed (default: 0).")
    demo_is.add_argument("--num-images", type=int, default=8, help="Number of images (default: 8).")
    demo_is.add_argument("--image-size", type=int, default=96, help="Square image size (default: 96).")
    demo_is.add_argument("--max-instances", type=int, default=2, help="Max instances per image (default: 2).")

    demo_cl = demo_sub.add_parser("continual", help="Toy continual-learning demo (requires torch; CPU OK).")
    demo_cl.add_argument("--output", default=None, help="Output JSON path or dir (default: runs/yolozu_demos/continual/...).")
    demo_cl.add_argument("--seed", type=int, default=0, help="Random seed (default: 0).")
    demo_cl.add_argument("--device", default="cpu", help="Torch device (default: cpu).")
    demo_cl.add_argument("--method", default="ewc_replay", choices=("naive", "ewc", "replay", "ewc_replay"))
    demo_cl.add_argument(
        "--methods",
        nargs="+",
        default=None,
        choices=("naive", "ewc", "replay", "ewc_replay"),
        help="Run multiple methods and write a suite report.",
    )
    demo_cl.add_argument(
        "--compare",
        action="store_true",
        help="Convenience flag: run all methods (naive/ewc/replay/ewc_replay) and write a suite report.",
    )
    demo_cl.add_argument(
        "--markdown",
        action="store_true",
        help="Also write a markdown summary table next to the JSON output (suite or single).",
    )
    demo_cl.add_argument("--steps-a", type=int, default=200, help="Training steps on domain A (default: 200).")
    demo_cl.add_argument("--steps-b", type=int, default=200, help="Training steps on domain B (default: 200).")
    demo_cl.add_argument("--batch-size", type=int, default=64, help="Batch size (default: 64).")
    demo_cl.add_argument("--hidden", type=int, default=32, help="Hidden units (default: 32).")
    demo_cl.add_argument("--lr", type=float, default=1e-2, help="Learning rate (default: 1e-2).")
    demo_cl.add_argument("--corr", type=float, default=2.0, help="Spurious correlation magnitude (default: 2.0).")
    demo_cl.add_argument("--noise", type=float, default=0.6, help="Feature noise std (default: 0.6).")
    demo_cl.add_argument("--n-train", type=int, default=4096, help="Train samples per domain (default: 4096).")
    demo_cl.add_argument("--n-eval", type=int, default=1024, help="Eval samples per domain (default: 1024).")
    demo_cl.add_argument("--ewc-lambda", type=float, default=20.0, help="EWC penalty weight (default: 20.0).")
    demo_cl.add_argument("--fisher-batches", type=int, default=64, help="Batches for Fisher estimate (default: 64).")
    demo_cl.add_argument("--replay-capacity", type=int, default=512, help="Replay buffer capacity (default: 512).")
    demo_cl.add_argument("--replay-k", type=int, default=64, help="Replay samples per step (default: 64).")

    args = parser.parse_args(argv)
    if args.command == "train":
        if not dev_enabled:
            raise SystemExit(
                "yolozu train is supported only from a source checkout (e.g. `pip install -e .`). "
                "Use `yolozu dev train <config>` when working in this repo."
            )
        config_path = Path(args.config)
        if not config_path.exists():
            raise SystemExit(f"config not found: {config_path}")
        return _cmd_train(config_path)
    if args.command == "test":
        if not dev_enabled:
            raise SystemExit(
                "yolozu test is supported only from a source checkout (e.g. `pip install -e .`). "
                "Use `yolozu dev test <config>` when working in this repo."
            )
        config_path = Path(args.config)
        if not config_path.exists():
            raise SystemExit(f"config not found: {config_path}")
        return _cmd_test(config_path)
    if args.command == "dev":
        if not dev_enabled:
            raise SystemExit(
                "yolozu dev is supported only from a source checkout (e.g. `pip install -e .`). "
                "The pip-installed CLI intentionally hides dev-only commands."
            )
        if args.dev_command == "train":
            config_path = Path(args.config)
            if not config_path.exists():
                raise SystemExit(f"config not found: {config_path}")
            return _cmd_train(config_path)
        if args.dev_command == "test":
            config_path = Path(args.config)
            if not config_path.exists():
                raise SystemExit(f"config not found: {config_path}")
            return _cmd_test(config_path)
        raise SystemExit("unknown dev command")
    if args.command == "doctor":
        return _cmd_doctor(str(args.output))
    if args.command == "export":
        return _cmd_export(args)
    if args.command == "predict-images":
        return _cmd_predict_images(args)
    if args.command == "eval-coco":
        return _cmd_eval_coco(args)
    if args.command == "parity":
        return _cmd_parity(args)
    if args.command == "validate":
        return _cmd_validate(args)
    if args.command == "eval-instance-seg":
        return _cmd_eval_instance_seg(args)
    if args.command == "onnxrt":
        if args.onnxrt_command == "export":
            return _cmd_onnxrt_export(args)
        raise SystemExit("unknown onnxrt command")
    if args.command == "resources":
        return _cmd_resources(args)
    if args.command == "demo":
        if args.demo_command == "instance-seg":
            from yolozu.demos.instance_seg import run_instance_seg_demo

            out = run_instance_seg_demo(
                run_dir=args.run_dir,
                seed=int(args.seed),
                num_images=int(args.num_images),
                image_size=int(args.image_size),
                max_instances=int(args.max_instances),
            )
            try:
                payload = json.loads(Path(out).read_text(encoding="utf-8"))
                res = payload.get("result", {})
                print(f"instance-seg demo: mAP50-95={res.get('map50_95'):.3f} mAP50={res.get('map50'):.3f}")
            except Exception:
                pass
            print(str(out))
            return 0

        if args.demo_command == "continual":
            from yolozu.demos.continual import (
                format_continual_demo_suite_markdown,
                run_continual_demo,
                run_continual_demo_suite,
            )

            methods = None
            if args.methods:
                methods = [str(m) for m in args.methods]
            elif args.compare:
                methods = ["naive", "ewc", "replay", "ewc_replay"]

            if methods and len(methods) > 1:
                out = run_continual_demo_suite(
                    methods=methods,
                    output=args.output,
                    seed=int(args.seed),
                    device=str(args.device),
                    steps_a=int(args.steps_a),
                    steps_b=int(args.steps_b),
                    batch_size=int(args.batch_size),
                    hidden=int(args.hidden),
                    lr=float(args.lr),
                    corr=float(args.corr),
                    noise=float(args.noise),
                    n_train=int(args.n_train),
                    n_eval=int(args.n_eval),
                    ewc_lambda=float(args.ewc_lambda),
                    fisher_batches=int(args.fisher_batches),
                    replay_capacity=int(args.replay_capacity),
                    replay_k=int(args.replay_k),
                )
                try:
                    payload = json.loads(Path(out).read_text(encoding="utf-8"))
                    md = format_continual_demo_suite_markdown(payload)
                    print(md, end="")
                    if args.markdown:
                        md_path = Path(out).with_suffix(".md")
                        md_path.write_text(md, encoding="utf-8")
                        print(str(md_path))
                except Exception:
                    pass
                print(str(out))
                return 0

            method = str(args.method)
            if methods and len(methods) == 1:
                method = str(methods[0])

            out = run_continual_demo(
                output=args.output,
                seed=int(args.seed),
                device=str(args.device),
                method=method,
                steps_a=int(args.steps_a),
                steps_b=int(args.steps_b),
                batch_size=int(args.batch_size),
                hidden=int(args.hidden),
                lr=float(args.lr),
                corr=float(args.corr),
                noise=float(args.noise),
                n_train=int(args.n_train),
                n_eval=int(args.n_eval),
                ewc_lambda=float(args.ewc_lambda),
                fisher_batches=int(args.fisher_batches),
                replay_capacity=int(args.replay_capacity),
                replay_k=int(args.replay_k),
            )
            try:
                payload = json.loads(Path(out).read_text(encoding="utf-8"))
                metrics = payload.get("metrics", {})
                a = metrics.get("after_task_a", {})
                b = metrics.get("after_task_b", {})
                forgetting = metrics.get("forgetting_acc_a")
                gain = metrics.get("gain_acc_b")
                print(
                    f"continual demo ({method}): "
                    f"accA {a.get('acc_a'):.3f}→{b.get('acc_a'):.3f} "
                    f"accB {a.get('acc_b'):.3f}→{b.get('acc_b'):.3f} "
                    f"forget={forgetting:.3f} gain={gain:.3f}"
                )
                if args.markdown:
                    md = format_continual_demo_suite_markdown({"runs": [{"method": method, "metrics": metrics}]})
                    md_path = Path(out).with_suffix(".md")
                    md_path.write_text(md, encoding="utf-8")
                    print(str(md_path))
            except Exception:
                pass
            print(str(out))
            return 0

    raise SystemExit("unknown command")


if __name__ == "__main__":
    raise SystemExit(main())
